#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 15:40:26 2020

@author: villa
"""

import numpy as np
import os
import os.path as op
import json
from monty.json import jsanitize,MontyEncoder, MontyDecoder


def get_object_feature(obj,feature):
    """
    Get value of attribute or method of a generic Object.
    If feature is a single method only the string with the method's name is required.
    If the target feature is stored in a dictionary (or dict of dictionaries), a list of
    this format needs to be provided: ["method_name",key1,key2,...]. 
        
    This will identify the value of Object.method[key1][key2][...] .

    Parameters
    ----------
    obj : object
        Generic object.
    feature : str or list
        Method or attribute of class for which the value is needed.
    
    Returns
    -------
    Object feature.

    """
    if isinstance(feature,list):
        method = feature[0]
        try:
            attr = getattr(obj,method) ()
        except:
            attr = getattr(obj,method)                
        for k in feature[1:]:
            try:
                v = attr[k]
            except:
                v = None
            if isinstance(v,dict):
                if feature.index(k) + 1 == len(feature):
                    return v
                else:
                    attr = v
            else:
                return v
            
    else:
        met = feature
        try:
            attr = getattr(obj,met) ()
        except:
            attr = getattr(obj,met)
        return attr


def get_object_from_json(object_class,path_or_string):
    """
    Build class object from json file or string. The class must posses the 'from_dict' method.

    Parameters
    ----------
    object_class : class
        Class of the object to decode.
    path_or_string : str
        If an existing path to a file is given the object is constructed reading the json file.
        Otherwise it will be read as a string.

    Returns
    -------
    Decoded object.

    """
    if op.isfile(path_or_string):
        with open(path_or_string) as file:
            d = json.load(file)
    else:
        d = json.loads(path_or_string)

    return object_class.from_dict(d)


def save_object_as_json(object,path,sanitize=False,cls=MontyEncoder):
    """
    Save class object as json string or file. The class must posses the `as_dict` method.

    Parameters
    ----------
    object: object
        Generic object. Must posses the `as_dict` method.
    path : str
        Path to the destination file.  If None a string is exported.

    Returns
    -------
    d : str
        If path is not set a string is returned.

    """
    d = object.as_dict()
    if sanitize:
        d = jsanitize(d)
    if path:
        with open(path,'w') as file:
            json.dump(d,file,cls=cls)
        return
    else:
        return json.dumps(d,cls=cls) 


def select_objects(objects,mode='and',exclude=False,functions=None,**kwargs):
    """
    Select objects from a list based on different criteria. Returns a list of objects.

    Parameters
    ----------
    objects : list
        List of objects.
    mode : str
        Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
    exclude : bool
        Exclude the entries satisfying the criteria instead of selecting them. The default is False.
    functions : list
        Functions containing criteria. The functions must take the object as the argument and
        return a bool. 
    kwargs : dict
        Keys are methods/attributes the objects have. Values contain the criteria. 
        To address more than one condition relative to the same attribute,
        use lists or tuples (e.g. charge=[0,1]).

    Returns
    -------
    output_objects : list
        List with selected objects.

    """    
    selected_objects = []
    entered_selection = False
    filtered_objects = objects.copy()
    if functions:
        for func in functions:
            if func:
                entered_selection = True
                if mode=='and':
                    if selected_objects:
                        filtered_objects = selected_objects.copy()
                        selected_objects = []
                for obj in filtered_objects:
                    if func(obj) == True:
                        if obj not in selected_objects:
                            selected_objects.append(obj)
    
    for key in kwargs:
        if mode=='and':
            if selected_objects or entered_selection is True:
                filtered_objects = selected_objects.copy()
                selected_objects = []
        for obj in filtered_objects:
            feature = get_object_feature(obj,key)
            if type(kwargs[key]) in [list,tuple]:
                for value in kwargs[key]:
                    if feature == value:
                        if obj not in selected_objects:
                            selected_objects.append(obj)
            elif feature == kwargs[key]:
                if obj not in selected_objects:
                    selected_objects.append(obj)  
    
    output_objects = []
    for obj in objects:
        if exclude:
            if obj not in selected_objects:
                output_objects.append(obj)
        else:
            if obj in selected_objects:
                output_objects.append(obj)    
    
    return output_objects



def sort_objects(objects,features,reverse=False):
    """
    Sort objects based on a list of features (attributes or methods of the objects, or functions)

    Parameters
    ----------
    objects : list
        List of objects to sort.
    features : list
        List of features (see get_object_feature).
    reverse : bool
        Reverse order.

    Returns
    -------
    sorted_objects : list
        Sorted objects.

    """
    def criteria(obj):
        criteria = []
        for feature in features:
            if callable(feature):
                criteria.append(feature(obj))
            else:
                criteria.append(get_object_feature(obj,feature))
        return criteria
    sort_function = lambda obj : criteria(obj)
    sorted_objects = sorted(objects,key=sort_function,reverse=reverse)
    
    return sorted_objects


from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition

def convert_conc_from_weight_to_cm3(c_weight,target_el,composition,bulk_volume=None,bulk_structure=None):
    """
    Convert concentration of dopand from weight % (common experimental data) to cm^-3.

    Parameters
    ----------
    c_weight : float
        Defect concentration in weight %.
    target_el : str
        Symbol of target element.
    composition : Composition
        Composition of the material, uses only list of elements.
    bulk_volume : float
        Volume of bulk cell in A°^3.
    bulk_structure : Structure
        If `bulk_volume` is not provided, use the Structure object of
        bulk material to optain the cell volume.
    
    Returns
    -------
    conc : float
        Concentration in cm-3^.

    """
    if type(composition) == str:
        composition = Composition(composition)

    if not bulk_volume and not bulk_structure:
        raise ValueError('You must provide either bulk_volume or bulk_structure')
    
    sum_MM = sum([el.atomic_mass for el in composition.elements])
    r_MM = Element(target_el).atomic_mass / sum_MM
    bulk_volume = bulk_volume or bulk_structure.lattice.volume
    conc = c_weight/r_MM * 1/bulk_volume * 1e24 *0.01
    return conc


def format_composition(string,all_math=False):
    """
    Format composition with latex syntax.
    """    
    def write(newc,new_string,skip,all_math):
        if not skip:
            if all_math:
                    new_string.append('_{%s}'%newc)
            else:
                new_string.append('$_{%s}$'%newc)
        return new_string
            
    skip = False
    new_string = []
    if all_math:
        new_string.append('$')
    sp = list(string)
    
    for c in sp:        
        if c.isdigit():
            try:
                nextc = sp[sp.index(c)+1]
            except:
                nextc = ''
            if nextc.isdigit():
                newc = c + nextc
                new_string = write(newc,new_string,skip,all_math)
                skip = True
            else:
                newc = c
                new_string = write(newc,new_string,skip,all_math)
                skip = False
        else:
            new_string.append(c)
    if all_math:
        new_string.append('$')       
    new_string = ''.join(new_string)
    
    return new_string


def get_charge_from_computed_entry(entry):
    """
    Get charge of vasp calculation from ComputedEntry object. 
    Subtracts the total valence electrons from potcar to 'NELECT' value
    """
    from pymatgen.io.vasp.inputs import Potcar

    potcar_symbols_bulk = [d['titel'].split(' ')[1] for d in entry.parameters['potcar_spec']]
    potcar = Potcar(potcar_symbols_bulk)
    charge = 0
    if 'parameters' in entry.data.keys():
        parameters = entry.data['parameters']
    else:
        raise ValueError('"parameters" need to be present in ComputedEntry data')
    if 'NELECT' in parameters.keys():
        nelect = parameters['NELECT']
        val = {}
        for p in potcar:
            val[p.element] = p.nelectrons
        neutral = sum([ val[el.symbol]*coeff 
                       for el,coeff in entry.structure.composition.items()])
        charge = neutral - nelect
    if not isinstance(charge,int):
        charge = np.around(charge,decimals=1)
        
    return charge


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 22 17:01:20 2023

@author: villa
"""
import os.path as op

from pymatgen.core.structure import Structure
from pymatgen.core.composition import Composition
from pymatgen.io.vasp.outputs import Vasprun

from defermi.structure import create_def_structure_for_visualization, defect_finder
from defermi.tools.utils import get_object_from_json

from defermi.testing.core import DefermiTest


class TestDefectFinder(DefermiTest):

    def setUp(self):
        self.structure_bulk = get_object_from_json(Structure,self.get_testfile_path('structure_bulk.json'))
        self.structure_vac = get_object_from_json(Structure,self.get_testfile_path('structure_vac.json'))
        self.structure_sub = get_object_from_json(Structure,self.get_testfile_path('structure_sub.json'))
        self.structure_int = get_object_from_json(Structure,self.get_testfile_path('structure_int.json'))
        self.structure_comp = get_object_from_json(Structure,self.get_testfile_path('structure_comp.json'))

    def test_finder_vacancy(self):
        vac = defect_finder(self.structure_vac,self.structure_bulk)
        assert vac.type == 'Vacancy'
        assert vac.defect_site_index == 0
        assert vac.specie == 'Si'
        
    def test_finder_substitution(self):
        sub = defect_finder(self.structure_sub,self.structure_bulk)
        assert sub.type == 'Substitution'
        assert sub.defect_site_index == 0
        assert sub.specie == 'P'
        
    def test_finder_interstitial(self):
        inter = defect_finder(self.structure_int,self.structure_bulk)
        assert inter.type == 'Interstitial'
        assert inter.defect_site_index == 54
        assert inter.specie == 'Si'
        
    def test_finder_complex(self):
        comp = defect_finder(self.structure_comp,self.structure_bulk)
        assert comp.type == 'DefectComplex'
        assert comp.defect_names == ['Sub_B_on_Si', 'Sub_P_on_Si']
        assert comp.delta_atoms == {'B': 1, 'Si': -2, 'P': 1}

    def test_finder_max_defects(self):
        structure_problematic = Vasprun(
            self.get_testfile_path('SiO2-defects/Defects/vacancies/Vac_O/q2/2-PBE-OPT/vasprun.xml'),
            parse_dos=False,parse_eigen=False,parse_potcar_file=False).final_structure
        structure_bulk = Vasprun(
            self.get_testfile_path('SiO2-defects/Bulk-2x2x2-supercell/vasprun.xml'),
            parse_dos=False,parse_eigen=False,parse_potcar_file=False).final_structure
        defect = defect_finder(
                            structure_defect=structure_problematic,
                            structure_bulk=structure_bulk,
                            tol=1e-02,
                            max_number_of_defects=1)
        assert defect.type == 'Vacancy'
        assert defect.specie == 'O'
        assert defect.defect_site_index == 24

        
    def test_create_def_structure_for_visualization(self):
        structure_vis = create_def_structure_for_visualization(self.structure_vac, self.structure_bulk,sort_to_bulk=True)
        assert structure_vis.composition == Composition('Si53P1')
        assert structure_vis[0].specie.symbol == 'P'
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 10:52:08 2020

@author: villa
"""

import numpy as np
import os.path as op
import os
import collections
from scipy.spatial import KDTree
from ase.visualize import view
from pymatgen.util.coord import pbc_shortest_vectors
from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Structure
from pymatgen.core.trajectory import Trajectory


def _get_distance_vector_and_image(lattice,frac_coords1,frac_coords2,jimage=None):
    """
    Same as pymatgen.core.Lattice.get_distance_and_image but returns 
    the distance vetor instead of the norm.
    Gets distance between two frac_coords assuming periodic boundary
    conditions. If the index jimage is not specified it selects the j
    image nearest to the i atom and returns the distance and jimage
    indices in terms of lattice vector translations. If the index jimage
    is specified it returns the distance between the frac_coords1 and
    the specified jimage of frac_coords2, and the given jimage is also
    returned.

    Parameters
    ----------
    frac_coords1 : np.array
        Reference fractional coords to get distance from.
    frac_coords2 : np.array
        fractional coords to get distance from.
    jimage : np.array
        Specific periodic image in terms of
        lattice translations, e.g., [1,0,0] implies to take periodic
        image that is one a-lattice vector away. If jimage is None,
        the image that is nearest to the site is found.

    Returns
    -------
    mapped_vec, jimage : tuple
        Distance and periodic lattice translations
        of the other site for which the distance applies. This means that
        the distance between frac_coords1 and (jimage + frac_coords2) is
        equal to distance.
    """
    if jimage is None:
        v, d2 = pbc_shortest_vectors(lattice, frac_coords1, frac_coords2, return_d2=True)
        fc = lattice.get_fractional_coords(v[0][0]) + frac_coords1 - frac_coords2  # type: ignore
        fc = np.array(np.round(fc), dtype=int)
        return v, fc

    jimage = np.array(jimage)
    mapped_vec = lattice.get_cartesian_coords(jimage + frac_coords2 - frac_coords1)  # type: ignore
    return mapped_vec, jimage  # type: ignore


def deform_lattice(structure,stdev=0.03):
    """
    Deform lattice vectors in structure. Also changes the angles of lattice vectors (off diagonal components).
    The function "apply_strain" from the Structure class only changes the base vectors' size but not the angles.

    Parameters
    ----------
    structure : Structure
        Structure object.
    stdev : float
        Standard deviation.

    Returns
    -------
    structure : Structure
        Structure with deformed lattice.
    """
    def get_random_strain(stdev):
        D = np.eye(3) #strain matrix
        for i in range(3):
            D[i, i] += np.random.uniform(-stdev, stdev)
            for j in range(i + 1, 3):
                D[i, j] += np.random.uniform(-stdev, stdev) / 4
                D[j, i] = D[i, j]
        return D
    strain_matrix = get_random_strain(stdev=stdev)
    atoms = structure.to_ase_atoms()
    new_cell = atoms.get_cell().dot(strain_matrix)
    atoms.set_cell(new_cell,scale_atoms=True)
    return Structure.from_ase_atoms(atoms)


def get_distance_vector(site1,site2,jimage=None):
    """
    Get distance vector between two sites assuming periodic boundary conditions.
    Same as pymatgen.core.sites.Site.distance but returns a vector instead of norm.

    Parameters
    ----------
    site 1 : PeriodicSite
        PeriodicSite object
    site 2 : PeriodicSite
        PeriodicSite object
    jimage : np.array
        Specific periodic image in terms of lattice
        translations, e.g., [1,0,0] implies to take periodic image
        that is one a-lattice vector away. If jimage is None,
        the image that is nearest to the site is found.
    
    Returns
    -------
    distance : float
        Distance between the two sites in A°.

    """
    return _get_distance_vector_and_image(site1.lattice, site1.frac_coords, site2.frac_coords,jimage=jimage)[0]


def get_displacement_vectors(structures):
    """
    Get vectors in cartesian coords of site displacements w.r.t. the first structure.

    Parameters
    ----------
    structures : list
        list of Pymatgen Structure objects.

    Returns
    -------
    displacements : np.array
        Displacement vectors in cartesian coordinates.

    """
    traj = Trajectory.from_structures(structures,constant_lattice=True) #order of structures determines ref in traj
    traj.to_displacements()
    disp = traj.coords[1]
    return structures[0].lattice.get_cartesian_coords(disp)


def get_furthest_neighbors(structure, target_site, species):
    """
    Get the indexes of atomic sites that are furthest away from the target site
    and maximally spaced apart from each other, considering periodic boundary conditions.

    Parameters
    ----------
    structure : Structure
    target_site : (int or PeriodicSite)
        PeriodicSite object or index of the target site in the structure
    species : list 
        List of species to find.

    Returns
    -------
    selected_indices : list
        List of indices corresponding to the furthest sites of the given species
    """
    if isinstance(target_site, int):
        target_site = structure[target_site]

    selected_indices = []
    
    for specie in species:
        max_score = -1
        best_index = None

        for i, site in enumerate(structure):
            if site.species_string != specie or i in selected_indices:
                continue
            
            # Distance from target
            distance_to_target = structure.get_distance(structure.index(target_site),i)

            # Compute minimum distance to already selected sites
            if selected_indices:
                distances_to_selected = [structure.get_distance(j,i) for j in selected_indices]
                min_distance_to_selected = min(distances_to_selected)
            else:
                min_distance_to_selected = distance_to_target

            # Define score: prioritize sites far from target and far from other selections
            score = distance_to_target + min_distance_to_selected

            if score > max_score:
                max_score = score
                best_index = i
        
        if best_index is not None:
            selected_indices.append(best_index)

    return selected_indices


def is_site_in_structure(site,structure,tol=1e-03):
    """
    Check if Site is part of the Structure list. This function is needed because 
    sometimes doing a simple check ("site in structure") doesn't work. This function performes
    a check on the coordinates and the element on the site. Therefore it is more reliable.

    Parameters
    ----------
    site : Site
        PeriodicSite or Site object.
    structure : Structure
        Structure object.
    tol : float
        Tolerance for fractional coordinates. The default is 1e-03.

    Returns
    -------
    is_site_in_structure : bool
    index : (int)
        Index of site in structure in case site is_site_in_structure returns True
        If False index will be None.
    """
    is_site_in_structure,index = False,None
    check,index = is_site_in_structure_coords(site, structure,tol=tol)
    if check:
        s = structure[index]
        if site.specie.symbol == s.specie.symbol:
            is_site_in_structure =True
    
    return is_site_in_structure,index


def is_site_in_structure_coords(site, structure, tol=1e-3, return_distance=False):
    """
    Check if a site's coordinates are present in a structure, using periodic boundary conditions.

    Parameters
    ----------
    site : Site
        PeriodicSite or Site object.
    structure : Structure
        Structure object.
    tol : float,
        Tolerance for site comparison, normalized with respect to lattice size.
    return_distance : bool
        Return distance btw site coords and closest site in reference structure. 

    Returns
    -------
    is_site_in_structure_coords : bool
        Whether the site exists in the structure within tolerance.
    index : int or None
        Index of matching site in structure if found, otherwise None.
    distance : float
        Distance btw site coords and closest site in reference structure.
        Returned if return_distance is set to True. 
        
    """
    # Normalize tolerance with lattice vector size
    l = site.lattice
    tol = np.sqrt(l.a**2 + l.b**2 + l.c**2) * tol

    # Convert structure sites to fractional coordinates
    frac_coords_structure = np.array([s.frac_coords % 1 for s in structure]) # ensure no negative coords
    kdtree_structure = KDTree(frac_coords_structure, boxsize=1.0)  # Take periodicity into account

    # Query the KDTree for nearest neighbor
    dist, index = kdtree_structure.query(site.frac_coords, distance_upper_bound=tol)

    if return_distance:
        is_site = [None,None,dist]
    else:
        is_site = [None,None]

    if dist < tol:
        is_site[0], is_site[1] = True, index
    else:
        is_site[0], is_site[1] = False, None
    
    return is_site


def rattle_atoms(structure,stdev=0.05,seed=None):
    """
    Change randomly atomic positions. Uses the ASE.
    Similar to the method "perturb" from the Structure class.

    Parameters
    ----------
    structure : Structure
        Structure object.
    stdev : float
        Standard deviation.
    seed : int
        Seed for random number generation passed to the Atoms.rattle method.

    Returns
    -------
    structure : Structure
        Structure with rattled atoms.    
    """
    atoms = structure.to_ase_atoms()
    atoms.rattle(stdev=stdev,seed=seed)
    return Structure.from_ase_atoms(atoms)


def remove_oxidation_state_from_site(site):
    """
    Remove `pymatgen` oxidation state decoration from Site.
    """
    new_sp = collections.defaultdict(float)
    for el, occu in site.species.items():
        sym = el.symbol
        new_sp[Element(sym)] += occu
    site.species = Composition(new_sp)
    return


def sort_sites_to_ref_coords(structure,structure_ref,extra_sites=[],tol=1e-03,get_indexes=False):
    """
    Sort Sites of one structure to match the order of coordinates in a reference structure. 

    Parameters
    ----------
    structure : Structure
        Structure to sort.
    structure_ref : Structure
        Reference Structure.
    extra_sites : list
        Sites to append at the end of the structure.
    tol: float
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size.
    get_indexes : bool
        Get list of mapping indexes for target structure sites in reference structure.

    Returns
    -------
    new_structure : Structure
        Sorted Structure.
    indexes : list
        If get_indexes is True a tuple is returned. List of mapping indexes for target structure sites
        in reference structure.
    """
    df = structure
    bk = structure_ref
    indexes = []
    new_sites=[]
    for s in df:
        check,index = is_site_in_structure_coords(s,bk,tol=tol)
        if check:
            indexes.append(index)
    for w in range(0,len(bk)):
        new_sites.insert(indexes[w],df[w])
    for s in extra_sites:
        new_sites.append(s)
            
    new_structure = df.copy()
    new_structure._sites = new_sites 
    if get_indexes:
        return new_structure, indexes
    else:
        return new_structure


def view_structures_with_ase(structures):
    """
    Visualize a Structure object or a list of Structure objects with the ASE. 
    First the Structure objects are converted into an ase Atom object, then "view" is used to visualize them.
    """
    if type(structures) == list:
        atoms=[]
        for s in structures:
            atoms.append(s.to_ase_atoms())
    else:
        atoms = structures.to_ase_atoms()
    view(atoms)
    return


def write_extxyz_file(file,structure,structure_ref=None,displacements=False):
    """
    Write extxyz format. Displacements can be included for visualization. 

    Parameters
    ----------
    file : str
        Path to save extxyz file to.
    structure : Structure
        Structure to visualize.
    structure_ref : Structure
        Reference Structure if needed.
    displacements : bool
        Include displacement vectors.
    """
    atoms = structure.to_ase_atoms()
    if displacements:
        disp = get_displacement_vectors([structure_ref, structure])
        atoms.arrays.update({'disp':disp})
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    atoms.write(file,format='extxyz')
    return


def write_xdatcar_from_structures(structures,file='XDATCAR'):
    """
    Write XDATCAR file from a list of structures. The first structure determines the reference for the Trajectory object.

    Parameters
    ----------
    structures : list
        List of Structure objects.
    file : str
        Path to write XDATCAR to.
    """
    traj = Trajectory.from_structures(structures,constant_lattice=True) 
    if not op.exists(op.dirname(file)):
        os.makedirs(op.dirname(file))
    traj.write_Xdatcar(file)
    return
    
    
    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  4 17:10:46 2023

@author: villa
"""

import warnings
import numpy as np
import os.path as op
import os

from pymatgen.core.sites import PeriodicSite
from pymatgen.core.periodic_table import Element
from pymatgen.core.trajectory import Trajectory

from .tools.structure import is_site_in_structure_coords, sort_sites_to_ref_coords, write_extxyz_file
from .defects import Vacancy,Substitution,Interstitial,DefectComplex
from .generator import create_interstitials, create_vacancies, create_substitutions


    
def create_interstitial_structures(structure,elements,supercell_size=None,**kwargs):
    """
    Create interstitial structures based on Voronoi with `pymatgen`,
    staring from a bulk structure (unit cell or supercell).
    Uses `Interstitial` objects generated with the `generator` module.

    Parameters
    ----------
    structure : Structure
        Bulk structure.
    elements : list
        List of element symbols.
    supercell_size : int
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified.
    kwargs: 
        Kwargs to pass to VoronoiInterstitialGenerator:

        - clustering_tol: Tolerance for clustering the Voronoi nodes.
        - min_dist: Minimum distance between an interstitial and the nearest atom.
        - ltol: Tolerance for lattice matching.
        - stol: Tolerance for structure matching.
        - angle_tol: Angle tolerance for structure matching.
        - kwargs: Additional keyword arguments for the `TopographyAnalyzer` constructor.

    Returns
    -------
    structures : list
        List of interstitial structures.

    """
    defects = create_interstitials(
                                structure=structure,
                                elements=elements,
                                supercell_size=supercell_size,
                                **kwargs)
    structures = [df.generate_defect_structure() for df in defects]
    return structures
    

def create_substitution_structures(structure,elements_to_replace,supercell_size=None,**kwargs):
    """
    Create substitution structures for each non-equivalent site,
    staring from a bulk structure (unit cell or supercell).
    Uses `Substitution` objects generated with the `generator` module.

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements_to_replace : str
        Dict with element symbol of specie to be replaced as keys and element 
        symbol of the species to be replaced with as values ({'old_El':'new_El'}).
    supercell_size : int or numpy array
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. 
    kwargs : dict
    
        Kwargs to pass to SpaceGroupAnalyzer class:

        - symprec (float): 
            Tolerance for symmetry finding. Defaults to 0.01,
            which is fairly strict and works well for properly refined
            structures with atoms in the proper symmetry coordinates. For
            structures with slight deviations from their proper atomic
            positions (e.g., structures relaxed with electronic structure
            codes), a looser tolerance of 0.1 (the value used in Materials
            Project) is often needed.

        - angle_tolerance (float): 
            Angle tolerance for symmetry finding. Defaults to 5 degrees.


    Returns
    -------
    structures : list
        List of substitution structures.

    """
    defects = create_substitutions(
                                structure=structure,
                                elements_to_replace=elements_to_replace,
                                supercell_size=supercell_size,
                                **kwargs)
    structures = [df.generate_defect_structure() for df in defects]
    return structures


def create_vacancy_structures(structure,elements=None,supercell_size=None,**kwargs):
    """
    Create Vacancy objects for each non-equivalent site,
    staring from a bulk structure (unit cell or supercell).
    Uses `Vacancy` objects generated with the `generator` module.

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements : str
        Symbol of the elements for which vacancies are needed.
        If None all of the elements are considered.
    supercell_size : int or numpy array
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. 
    kwargs : dict
    
        Kwargs to pass to SpaceGroupAnalyzer class:

        - symprec (float): 
            Tolerance for symmetry finding. Defaults to 0.01,
            which is fairly strict and works well for properly refined
            structures with atoms in the proper symmetry coordinates. For
            structures with slight deviations from their proper atomic
            positions (e.g., structures relaxed with electronic structure
            codes), a looser tolerance of 0.1 (the value used in Materials
            Project) is often needed.

        - angle_tolerance (float): 
            Angle tolerance for symmetry finding. Defaults to 5 degrees.
            

    Returns
    -------
    structures : (list)
        List of vacancy structures.
    """
    defects = create_vacancies(
                            structure=structure,
                            elements=elements,
                            supercell_size=supercell_size,
                            **kwargs)
    structures = [df.generate_defect_structure() for df in defects]
    return structures

 
def create_def_structure_for_visualization(structure_defect,structure_bulk,defects=None,sort_to_bulk=False,tol=1e-03):
    """
    Create defect structure for visualization in OVITO. The vacancies are shown by inserting 
    in the vacant site the element of same row and next group on the periodic table.
    If sort_to_bulk is True the Sites are sorted to match the Bulk structure.

    Parameters
    ----------
    structure_defect : Structure
        Defect structure.
    structure_bulk : Structure
        Bulk structure.
    defects : list 
        List of defect objects. If None `defect_finder` is used.
    sort_to_bulk : bool
        Sort Sites of the defect structure to match the order of coordinates in the bulk structure
        (useful if the non-relaxed defect structure is not available). 
        If False only the dummy atoms are inserted and not further changes are made.
    tol: float
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.

    Returns
    -------
    new_structure : Structure 
        Structure with dummy atoms as vacancies and interstitials at the end of the list.
        The order of the sites follow the order of the Bulk structure.

    """
    df = structure_defect.copy()
    bk = structure_bulk.copy()
    extra_sites=[]
    if defects:
        dfs = defects
    else:
        df_found = defect_finder(df,bk,tol=tol)
        if df_found.type=='DefectComplex':
            dfs = df_found.defects
        else:
            dfs = [df_found]
   
    for d in dfs:
        dsite = d.site
        dtype = d.type
        if dtype == 'Vacancy':
            check,i = is_site_in_structure_coords(dsite,bk,tol=tol)
            el = dsite.specie
            species = Element.from_row_and_group(el.row, el.group+1)
            df.insert(idx=i,species=species,coords=dsite.frac_coords)
        elif dtype == 'Interstitial' and sort_to_bulk:
            extra_sites.append(dsite)

    # reorder to match bulk, useful if you don't have the non-relaxed defect structure
    if sort_to_bulk:
        new_structure = sort_sites_to_ref_coords(df, bk, extra_sites,tol=tol)
    # In this case only dummy atoms are inserted, no further changes
    else:
        new_structure = df.copy()
    return new_structure        


def defect_finder(
                structure_defect,
                structure_bulk,
                tol=1e-3,
                max_number_of_defects=None,
                verbose=False):
    """
    Find defects by comparing defect and bulk structures.

    Parameters
    ----------
    structure_defect : Structure
        Defect structure.
    structure_bulk : Structure
        Bulk structure.
    tol : float
        Tolerance for fractional coordinates comparison (default is 1e-3).
    max_number_of_defects : int
        Impose a max number of defects to be found. If set, defects
        are ranked based on the coordinate distance btw defect and bulk sites
        (descending order), the first max_number_of_defects in the list 
        are given as the output (as single defect or defect complex).
    verbose : bool
        Print output.

    Returns
    -------
    Defect object (Vacancy, Substitution, Interstitial, DefectComplex). Not implemented for Polaron.

    """
    defects = []
    # Identify missing (vacancies) and additional (interstitials) sites
    matched_indices = set()
    for site in structure_defect:
        check, index = is_site_in_structure_coords(site,structure_bulk,tol=tol,return_distance=False)

        if check:
            matched_indices.add(index)  # Site exists in bulk, check for substitution
            if site.species != structure_bulk[index].species:
                defects.append(
                    Substitution(specie=site.specie.symbol,
                                defect_site=site,
                                bulk_structure=structure_bulk,
                                site_in_bulk=structure_bulk[index]))
        else:
            defects.append(Interstitial(specie=site.specie.symbol,
                                        defect_site=site,
                                        bulk_structure=structure_bulk))

    for j, site in enumerate(structure_bulk):
        if j not in matched_indices:
            defects.append(Vacancy(specie=site.specie.symbol,
                                        defect_site=site,
                                        bulk_structure=structure_bulk))

    if max_number_of_defects:
        filtered_defects = []
        defect_distances = []
        for df in defects:
            if df.type == 'Vacancy':
                check, i, d = is_site_in_structure_coords(df.site,structure_defect,tol=tol,return_distance=True)
                defect_distances.append((df,d))
            else:
                check, i, d = is_site_in_structure_coords(df.site,structure_bulk,tol=tol,return_distance=True)
                defect_distances.append((df,d))
        sorted_defect_distances = defect_distances.sort(key = lambda x : x[1], reverse=True)
        filtered_defects = [defect_distances[i][0] for i in range(max_number_of_defects)]
    else:
        filtered_defects = defects

    if len(filtered_defects) > 1:
        if len(filtered_defects) > 3:
            warnings.warn("More than 3 defects found, if not desired try to adjust the tolerance parameter ")
        defect = DefectComplex(filtered_defects, bulk_structure=structure_bulk)
    elif len(filtered_defects) == 1:
        defect = filtered_defects[0]
    else:
        warnings.warn("No defect has been found. Try to adjust the tolerance parameter.", UserWarning)
        defect = None
    
    if verbose:
        print(f'Defect automatically identified for defective structure with composition {structure_defect.composition}: \n {defect}')

    return defect


def get_trajectory_for_visualization(structure_defect,structure_bulk,defects=None,tol=1e-03,file=None):
    """
    Create trajectory from defect and bulk structures for visualization. 
    The vacancies are shown by inserting in the vacant site the element of same row and next group on the periodic table.

    Parameters
    ----------
    structure_defect : Structure
        Defect structure.
    structure_bulk : Structure
        Bulk structure.
    defects : list 
        List of Defect objects. If None `defect_finder` is used.
    tol: float
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    file : str
        File to save XDATCAR. 
    
    Returns
    -------
    new_structure : Structure
        Structure with dummy atoms as vacancies and interstitials in the bottom.
        The order of the Sites follow the order of the Bulk structure.

    """
    sb = structure_bulk
    dummy = create_def_structure_for_visualization(structure_defect, structure_bulk,defects,sort_to_bulk=True,tol=tol)
    traj = Trajectory.from_structures([dummy,sb],constant_lattice=True)     
    if file:
        if not op.exists(op.dirname(file)):
            os.makedirs(op.dirname(file))
        traj.write_Xdatcar(file)
    return traj
        
        
def write_extxyz_for_visualization(file,structure_defect,structure_bulk,defects=None,tol=1e-03): 
    """
    Write extxyz file for visualization. The displacements w.r.t the bulk structure are included.
    The vacancies are shown by inserting in the vacant site the element of same row and next group on the periodic table.
    
    Parameters
    ----------
    file : str
        Path to save file. 
    structure_defect : Structure
        Defect structure.
    structure_bulk : Structure
        Bulk structure.
    defects : list 
        List of Defect objects. If None `defect_finder` is used.
    tol: float
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    
    """
    sb = structure_bulk
    dummy = create_def_structure_for_visualization(structure_defect, structure_bulk,defects,sort_to_bulk=True,tol=tol)
    write_extxyz_file(file, dummy, sb,displacements=True)
    return


#!/usr/bin/env python

import warnings




class  MPDatabase:
    
    from mp_api.client import MPRester # only import within this class

    def __init__(self,mp_id=None,API_KEY=None):
        """
        Class to retrieve data from Materials Project database

        Parameters
        ----------
        mp_id : str
            Materials-ID.
        API_KEY : str
            API_KEY for MAterials Project database. If None the default key from
            the configuration file for `pymatgen` is used.
        """
        
        self.mp_id = mp_id if mp_id else None
        self.API_KEY = API_KEY
    
    @property
    def mp_rester(self):
        return self.MPRester(self.API_KEY)
        
        
    def get_entries(self,
                    chemsys_formula_mpids,
                    compatible_only=True,
                    property_data=None,
                    conventional_unit_cell=False,
                    sort_by_e_above_hull=True):
        """
        Get a list of ComputedEntries or ComputedStructureEntries corresponding
        to a chemical system, formula, or materials_id or full criteria.

        Parameters
        ----------
        chemsys_formula_id_criteria : str or dict
            A chemical system (e.g., Li-Fe-O), or formula (e.g., Fe2O3) 
            or materials_id (e.g., mp-1234) or full Mongo-style dict criteria.
        compatible_only : bool
            Whether to return only "compatible"
            entries. Compatible entries are entries that have been
            processed using the MaterialsProject2020Compatibility class,
            which performs adjustments to allow mixing of GGA and GGA+U
            calculations for more accurate phase diagrams and reaction
            energies.
        property_data : list
            Specify additional properties to include in
            entry.data. If None, no data. Should be a subset of
            supported_properties.
        conventional_unit_cell : bool
            Whether to get the standard conventional unit cell.

        Returns
        -------
        entries : list
            List of ComputedStructureEntry objects.

        """
        with self.MPRester(self.API_KEY) as mpr:
            entries = mpr.get_entries(
                                chemsys_formula_mpids=chemsys_formula_mpids,
                                compatible_only=compatible_only,
                                property_data=property_data,
                                conventional_unit_cell=conventional_unit_cell)
        return entries
    
    
    def get_entries_from_compositions(self,
                                    compositions,
                                    compatible_only=True,
                                    property_data=None,
                                    conventional_unit_cell=False):
        """
        Get a dictionary with compositions (strings) as keys and a list of ComputedEntries 
        or ComputedStructureEntries as values.

        Parameters
        ----------
        compositions : list
            List of strings with compositions.
        compatible_only : bool
            Whether to return only "compatible"
            entries. Compatible entries are entries that have been
            processed using the MaterialsProject2020Compatibility class,
            which performs adjustments to allow mixing of GGA and GGA+U
            calculations for more accurate phase diagrams and reaction
            energies.
        property_data : list
            Specify additional properties to include in
            entry.data. If None, no data. Should be a subset of
            supported_properties.
        conventional_unit_cell : bool
            Whether to get the standard conventional unit cell.

        Returns
        -------
        entries : list
            List of ComputedStructureEntry objects.

        """
        entries_dict = {}
        for comp in compositions:
            entries = self.get_entries(
                                    chemsys_formula_mpids=comp,
                                    compatible_only=compatible_only,
                                    property_data=property_data,
                                    conventional_unit_cell=conventional_unit_cell)
            entries_dict[comp] = entries
        
        return entries_dict
    

    def get_phase_diagram_from_chemsys(self,chemsys,thermo_type='GGA_GGA+U', **kwargs):
        """
        Pull `PhaseDiagram` object from MP database given a chemical system (eg. "Li-Nb-O").

        Parameters
        ----------
        chemsys : str
            Chemical system (eg. "Li-Nb-O").
        thermo_type : str
            Thermo types to return data for (e.g. "GGA_GGA+U").
            Check `mp_api` docs for information on thermo types.
        kwargs : dict
            Kwargs to pass to `MPRester().materials.thermo.get_phase_diagram_from_chemsys`.

        Returns
        -------
        pd : PhaseDiagram
            `PhaseDiagram` object.
        """
        with self.MPRester(self.API_KEY) as mpr:
            pd = mpr.materials.thermo.get_phase_diagram_from_chemsys(
                                                                chemsys=chemsys,
                                                                thermo_type=thermo_type,
                                                                **kwargs)
        return pd
        

    def get_stable_energy_pfu_from_composition(self,composition,thermo_types=['GGA_GGA+U'],**kwargs):
        """
        Pull from MP database for a target composition the energy per formula 
        unit (pfu) in eV of the entry with E above Convex Hull = 0 eV.

        Parameters
        ----------
        composition : str
            Target composition (eg. "SrTiO3").
        thermo_type : str
            Thermo types to return data for (e.g. "GGA_GGA+U").
            Check `mp_api` docs for information on thermo types.
        kwargs : dict
            Kwargs to pass to `MPRester().materials.thermo.get_phase_diagram_from_chemsys`.

        Returns
        -------
        energy_pfu : float
            Energy per formula unit in eV.

        """
        with self.MPRester(self.API_KEY) as mpr:
            docs = mpr.materials.thermo.search(formula=composition,energy_above_hull=(0,0),thermo_types=thermo_types,**kwargs)
        
        if len(docs) > 1:
            warnings.warn('Search returned more than one entry with E above hull = 0 eV, check manually. Returning the first entry...')
        
        entry = docs[0]
        nfu = entry.composition.get_reduced_composition_and_factor()[1]
        energy_pfu = entry.energy_per_atom * (entry.composition.num_atoms/nfu)
        return energy_pfu
        

    def get_structure(self,final=True,conventional_unit_cell=False):
        """
        Get a Structure corresponding to a material ID.

        Parameters
        ----------
        material_id : str
            Materials Project material_id (a string, e.g., mp-1234).
        final : bool
            Whether to get the final structure, or the initial
            (pre-relaxation) structure.
        conventional_unit_cell : bool
            Whether to get the standard conventional unit cell

        Returns
        -------
        structure : Structure
            Structure object.
            
        """
        with self.MPRester(self.API_KEY) as mpr:
            structure = mpr.get_structure_by_material_id(self.mp_id,final=final,conventional_unit_cell=conventional_unit_cell)
        return structure

    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  2 16:39:01 2025

@author: lorenzo
"""

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt
from pymatgen.core.periodic_table import Element
from pymatgen.core.composition import Composition
from pymatgen.symmetry.groups import SpaceGroup
from pymatgen.analysis.phase_diagram import GrandPotentialPhaseDiagram, PDPlotter

from ..tools.utils import format_composition
from .core import Chempots


def _get_composition_object(comp):
    if type(comp) == str:
        return Composition(comp)
    elif type(comp) == Composition:
        return comp
    else:
        raise ValueError('Composition needs to be str or pymatgen Composition')
        

class PDHandler:
    
    def __init__(self,phase_diagram):
        """
        Class to generate and handle Pymatgen phase diagram more rapidly.

        Parameters
        ----------
        phase_diagram : PhaseDiagram
            Pymatgen PhaseDiagram object

        """
        self.pd = phase_diagram 
        self.mu_refs = self.get_chempots_reference()


    def calculate_single_chempot(self,comp,chempots_ref):
        """
        Calculate referenced chemical potential in a given composition and given
        the chemical potentials of the other elements.

        Parameters
        ----------
        comp : str or Composition
            Compositions of the phase.
        chempots_ref : Dict
            Dictionary with element symbols as keys and respective chemical potential as value ({el:chempot}).
            The chemical potentials used here are the ones relative to the reference (delta_mu).

        Returns
        -------
        mu : float
            Chemical potential.

        """        
        comp = _get_composition_object(comp)
        form_energy = self.get_formation_energy_from_stable_comp(comp)
        mu = form_energy
        for el,coeff in comp.items():
            if el.symbol in chempots_ref:
                mu += -1*coeff * chempots_ref[el.symbol]
            else:
                factor = coeff
        
        return mu/factor


    def get_all_boundaries_chempots(self,comp):
        """
        Get chemical potentials at the corners of the stability ta given composition.

        Parameters
        ----------
        comp : str or Composition

        Returns
        -------
        chempots : Chempots
            Chempots object.

        """
        comp = _get_composition_object(comp)
        chempots_pmg = self.pd.get_all_chempots(comp)
        for r,mu in chempots_pmg.items():
            chempots_pmg[r] = {k:v.item() for k,v in mu.items()} # convert from np.float64
        chempots = {r:Chempots.from_pmg_elements(mu) for r,mu in chempots_pmg.items()}
        return chempots

        
    def get_chempots_reference(self):
        """
        Gets elemental reference compounds and respective e.p.a with Pymatgen el_ref attribute in PhaseDiagram class.

        Returns
        -------
        chempots : Chempots
            Chempots object.

        """
        chempots_ref = {}
        pd = self.pd
        for el in pd.el_refs:
            chempots_ref[el] = pd.el_refs[el].energy_per_atom
        chempots_ref = {k: v for k, v in sorted(chempots_ref.items(), key=lambda item: item[0])}
        return Chempots.from_pmg_elements(chempots_ref)
       

    def get_dataframe(self):
        """
        Generate pandas DataFrame with columns 'Composition, Structure, Formation energy'.
        To display a string for 'Structure' the entry needs to be a ComputedStructureEntry (see pymatgen docs).
        """
        phases = []
        for e in self.pd.stable_entries:
            d = {}
            d['Composition'] = format_composition(e.composition.reduced_formula)
            sg = SpaceGroup(e.structure.get_space_group_info()[0])
            crystal_system, sg_string = sg.crystal_system , sg.to_latex_string()
            if e.__class__.__name__ == 'ComputedStructureEntry':
                d['Structure'] = f'{crystal_system.capitalize()} ({sg_string})'
            else:
                d['Structure'] = None
            d['Formation energy p.a (eV)'] = np.around(self.pd.get_form_energy_per_atom(e),decimals=2)
            phases.append(d)
        df = DataFrame(phases)
        return df

        
    def get_entries_from_comp(self,comp):
       """
       Get a list of entries corrisponding to the target composition.
       
       Parameters
       ----------
       comp : str or Composition
            Composition.

       Returns
       -------
       target_entries : list
        List of Pymatgen PDEntry objects.

       """
       comp = _get_composition_object(comp)
       target_entries=[]
       pd = self.pd
       for e in pd.all_entries:
           if e.composition.reduced_composition == comp:
               target_entries.append(e)
       if target_entries != []:
           return target_entries
       else:
           raise ValueError('No entry has been found for target composition:%s' %comp.reduced_formula)


    def get_energies_from_comp(self,comp):
        """
        Get dictionary of energies for all entries of a given reduced composition.

        Parameters
        ----------
        comp : str or Composition
            Composition.

        Returns
        -------
        form_energies : dict
            Dictionary with PDEntry objects as keys and Energies as values.

        """
        comp = _get_composition_object(comp)
        pd = self.pd
        energies = {}
        for e in self.get_entries_from_comp(comp):
            factor = e.composition.get_reduced_composition_and_factor()[1]
            energies[e] = e.energy/factor
        return energies
            

    def get_energy_from_stable_comp(self,comp):
        """
        Get energy of a target stable composition.

        Parameters
        ----------
        comp : str or Composition
            Composition.

        Returns
        -------
        energy : float
            Energy in eV.

        """
        comp = _get_composition_object(comp)
        pd = self.pd
        entry = self.get_stable_entry_from_comp(comp)
        factor = entry.composition.get_reduced_composition_and_factor()[1]
        return entry.energy/factor


    def get_formation_energies_from_comp(self,comp):
        """
        Get dictionary of formation energies for all entries of a given reduced composition.

        Parameters
        ----------
        comp : str or Composition
            Composition.

        Returns
        -------
        form_energies : dict
            Dictionary with PDEntry objects as keys and formation energies in eV as values.
        
        """
        comp = _get_composition_object(comp)
        pd = self.pd
        form_energies = {}
        for e in self.get_entries_from_comp(comp):
            factor = e.composition.get_reduced_composition_and_factor()[1]
            form_energies[e] = pd.get_form_energy(e)/factor
        return form_energies
            

    def get_formation_energy_from_stable_comp(self,comp):
        """
        Get formation energy of a target stable composition.

        Parameters
        ----------
        comp : str or Composition
            Composition.

        Returns
        -------
        form_energy : float
            Formation energy in eV (float).

        """
        comp = _get_composition_object(comp)
        pd = self.pd
        entry = self.get_stable_entry_from_comp(comp)
        factor = entry.composition.get_reduced_composition_and_factor()[1]
        return pd.get_form_energy(entry)/factor
        

    def get_phase_boundaries_chempots(self,comp,chempot_ref):
        """
        Given a composition and a fixed chemical potential, this function analises the composition of the boundary phases
        and the associated chemical potentials at the boundaries. Only works for 3 component PD.

        Parameters
        ----------
        comp : str or Composition
            Composition of the phase you want to get the chemical potentials at the boundary.
        chempot_ref : dict
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}).
            The chemical potential here is the referenced value. 

        Returns
        -------
        chempots : dict
            Dictionary with compositions at the boundaries as keys and delta chemical potentials as value.

        """       
        comp = _get_composition_object(comp)
        chempots = {}
        comp1,comp2 = self.get_phase_boundaries_compositions(comp, chempot_ref)
        
        boundary = '-'.join([comp1.reduced_formula,comp.reduced_formula])
        chempots[boundary] = self.solve_phase_boundary_chempots(comp1, comp, chempot_ref)
        
        boundary = '-'.join([comp.reduced_formula,comp2.reduced_formula])
        chempots[boundary] = self.solve_phase_boundary_chempots(comp, comp2, chempot_ref)
        
        return chempots      
                  
    
    def get_phase_boundaries_compositions(self,comp,chempot_ref):
        """
        Get compositions of phases in boundary of stability with a target composition given a fixed chemical potential 
        on one component. Currently only works for 3-component PD (to check). 
        Used Pymatgen GrandPotentialPhaseDiagram class. The fixed chemical potential is the referenced value that is
        converted in the global value for the analysis with the GrandPotentialPhaseDiagram class.

        Parameters
        ----------
        comp : str or Composition
            Target composition for which you want to get the bounday phases.
        chempot_ref : dict
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}).
            The chemical potential is the referenced value

        Returns
        -------
        comp1,comp2 : (Composition objects)
            Compositions of the boundary phases given a fixed chemical potential for one element.

        """
        comp = _get_composition_object(comp)
        chempot_ref = Chempots(chempot_ref)
        fixed_chempot = chempot_ref.get_absolute(self.mu_refs).to_pmg_elements()
        
        entries = self.pd.all_entries
        gpd = GrandPotentialPhaseDiagram(entries, fixed_chempot)
        stable_entries = gpd.stable_entries
        comp_in_stable_entries = False
        for e in stable_entries:
            if e.original_comp.reduced_composition == comp:
                comp_in_stable_entries = True
        if comp_in_stable_entries == False:
            raise ValueError('Target composition %s is not a stable entry for fixed chemical potential: %s' %(comp.reduced_formula,fixed_chempot))
        
        el = comp.elements[0]
        x_target = comp.get_wt_fraction(el)
        x_max_left = 0
        x_min_right = 1
        for e in stable_entries:
            c = e.original_comp
            if c != comp:
                x = c.get_wt_fraction(el)
                if x < x_target and x >= x_max_left:
                    x_max_left = x
                    comp1 = c
                if x > x_target and x <= x_min_right:
                    x_min_right = x
                    comp2 = c
        return comp1.reduced_composition,comp2.reduced_composition   
                  
                    
    def solve_phase_boundary_chempots(self,comp1,comp2,chempot_ref):
        """
        Given a fixed chemical potential, gets the values of the remaining two chemical potentials
        in the boundary between two phases (region where the two phases coexist). Only works for 3-component PD (to check).

        Given a phase P1 (formula AxByOz) and a phase P2 (formula AiBjOk) the chemical potentials have to satisfy the conditions:

        - form_energy(P1) = x*mu(A) + y*mu(B) +z*mu(O)
        - form_energy(P2) = i*mu(A) + j*mu(B) +k*mu(O)
        
        From these conditions the values of mu(A) and mu(B) are determined given a fixed value of mu(O).
        All of the chemical potentials used here are delta_mu, i.e. relative to the elemental phase(delta_mu(O) = mu(O) - mu_ref(O))

        Parameters
        ----------
        comp1,comp2 : str or Composition
            Compositions of the two phases at the boundary.
        chempot_ref : dict
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}). The chemical potential
            used here is the one relative to the reference (delta_mu).

        Returns
        -------
        chempots_boundary : dict
            Dictionary of chemical potentials.

        """
        comp1 = _get_composition_object(comp1)
        comp2 = _get_composition_object(comp2)
        chempots_boundary ={}
        chempot_ref = Chempots(chempot_ref)
        mu_fixed = chempot_ref.to_pmg_elements()
        for el,chempot in mu_fixed.items():
            el_fixed, mu_fixed = el, chempot
        e1 = self.get_formation_energy_from_stable_comp(comp1)
        e2 = self.get_formation_energy_from_stable_comp(comp2)
        
        coeff1 = []
        coeff2 = []
        # order of variables (mu) will follow the order of self.chempots_reference which is alphabetically ordered
        mu_refs = self.mu_refs.to_pmg_elements()
        for el in mu_refs:
            if el != el_fixed:
                if el not in comp1:
                    coeff1.append(0)
                else:
                    coeff1.append(comp1[el])
                if el not in comp2:
                    coeff2.append(0)
                else:
                    coeff2.append(comp2[el])                
        a = np.array([coeff1,coeff2])
        
        const1 = e1 - comp1[el_fixed]*mu_fixed if el_fixed in comp1 else e1 
        const2 = e2 - comp2[el_fixed]*mu_fixed if el_fixed in comp2 else e2
        b = np.array([const1,const2])
        x = np.linalg.solve(a, b)
        # output will follow the order given in input
        counter = 0
        for el in mu_refs:
            if el != el_fixed:
                chempots_boundary[el] = x[counter]
                counter += 1
        chempots_boundary[el_fixed] = mu_fixed        
        
        return Chempots.from_pmg_elements(chempots_boundary) 


    def get_plot(self,**kwargs):
        """
        Get plot with Pymatgen
        """
        PDPlotter(self.pd,show_unstable=0,backend='matplotlib').get_plot(**kwargs)
        return plt


    def get_stability_diagram(self,elements,figsize=None):
        """
        Method to get stability diagram with 'get_chempot_range_map_plot' method in pymatgen.

        Parameters
        ----------
        elements : list
            List with strings of the elements to be used as free variables.
        size : tuple
            New size in inches.

        Returns
        -------
        plt : 
            Matplotlib object.

        """
        pd = self.pd
        elements = [Element(el) for el in elements]
        PDPlotter(pd).get_chempot_range_map_plot(elements)
        if figsize:
            fig = plt.gcf()
            fig.set_size_inches(figsize[0],figsize[1])
        
        return plt


    def get_stable_entry_from_comp(self,comp):
        """
        Get the PDEntry of the stable entry of a target composition.

        Parameters
        ----------
        comp : str or Composition
            Composition.

        Returns
        -------
        entry : PDEntry
            Pymatgen PDEntry object.

        """
        comp = _get_composition_object(comp)
        target_entry=None
        pd = self.pd
        for e in pd.stable_entries:
            if e.composition.reduced_composition == comp:
                target_entry = e
                break
        if target_entry is not None:
            return target_entry
        else:
            raise ValueError('No stable entry has been found for target composition:%s' %comp.reduced_formula)
            
            
            
class StabilityDiagram:
    
    def __init__(self,phase_diagram=None,size=1):
        """
        Class with tools to add features to default PD plots generated by Pymatgen.

        Parameters
        ----------
        phase_diagram : PhaseDiagram
            Pymatgen PhaseDiagram object.
        size : float
            Multiplier for the size of the objects added in the plot.

        """
        self.pd = phase_diagram if phase_diagram else None
        self.pdh = PDHandler(phase_diagram) if phase_diagram else None
        self.size = size
        
    
    def add_points(self,points,size=1,label_size=1,color=[],edgecolor='k',label_color='k',linewidths=3,**kwargs):
        """
        Add points to plot representing reservoirs.

        Parameters
        ----------
        points : dict
            Dictionary with points labels as keys and tuples or list with coordinates as values.
        size : float
            Float multiplier for points size. Default is 1, which would yield a default size of 450*self.size
        label_size : float
            Float multiplier for labels size. Default is 1, which would yield a default size of 30*self.size
        color : str
            Color of filling of points
        edgecolor : str 
            Color of point edge
        label_color : str
            Color of labels
        linewidths : int
            line width of point edge
        kwargs : dict
            kwargs to pass to matplotlib `scatter` function.

        Returns
        -------
        plt : matplotlib
            Matplotlib object.

        """
        for p in points:
            plt.scatter(points[p][0],points[p][1], color=color, edgecolor=edgecolor, linewidths=linewidths, s=450*self.size*size,**kwargs)
            plt.text(points[p][0]+(0.1/self.size*label_size),points[p][1],p,size=30*self.size*label_size,color=label_color)
        return plt
    
    
    def add_constant_chempot_line(self, comp, variable_element, chempots_ref,**kwargs):
        """
        Add line of constant chemical potential (at a given composition) to the plot. Only works for 3 component PD.

        Parameters
        ----------
        comp : str or Composition
            Composition of the phase.
        variable_element : str
            Element chosen as indipendent variable.
        chempots_ref : dict
            Dictionary with fixed chemical potentials (values relative to reference phase). The format is {Element:chempot}.
        **kwargs : 
            kwargs passed to Matplotlib plot function.

        Returns
        -------
        plt : matlotlib
            Matplotlib object.

        """
        comp = _get_composition_object(comp)
        axes = plt.gca()
        xlim , ylim = axes.get_xlim() , axes.get_ylim()
        plt.xlim(xlim)
        plt.ylim(ylim)
        mu = np.arange(xlim[0]-1,xlim[1]+1,0.01)
        plt.plot(mu,self.constant_chempot_line(mu,comp,variable_element,chempots_ref),
                 linewidth= 4.5*self.size , **kwargs)
        return plt
    

    def add_heatmap(self,comp,elements,cbar_label='$\Delta\mu_{O}$',cbar_values=True,**kwargs):
        """
        Add heatmap that shows the value of the last chemical potential based on the values of the other two "free" 
        chemical potentials and the composition of interest. Currently works only for 3 component PDs.

        Parameters
        ----------
        comp : str or Composition
            Composition of interest to compute the chemical potential.
        elements : list
            List of strings with elements with free chemical potentials.
        cbar_label : str
            String with label of the colormap. The default is ''.
        cbar_values : tuple or bool
            Show max e min chempot values on colorbar. If tuple the values are used, if not the 
            minimum chempot and 0 are used. The default is True.
        **kwargs : dict
            kwargs for "pcolormesh" function.

        Returns
        -------
        plt : matplotlib
            Matplotlib object.

        """        
        comp = _get_composition_object(comp)
        el1,el2 = elements  
        
        def f(mu1,mu2):            
            return self.pdh.calculate_single_chempot(comp,{el1:mu1,el2:mu2})
        
        axes = plt.gca()
        xlim , ylim = axes.get_xlim() , axes.get_ylim()
        npoints = 100
        x = np.arange(xlim[0],xlim[1]+0.1,abs(xlim[1]+0.1-xlim[0])/npoints)
        y = np.arange(ylim[0],ylim[1]+0.1,abs(ylim[1]+0.1-ylim[0])/npoints)   
        
        X,Y = np.meshgrid(x,y)
        Z = f(X,Y)

        plt.pcolormesh(X,Y,Z,vmax=0,shading='auto',**kwargs)

        cbar = plt.colorbar()
       # cbar.ax.tick_params(labelsize='xx-large')
        if cbar_values:
            if isinstance(cbar_values,tuple):
                cbar_min,cbar_max = cbar_values[0], cbar_values[1]
            else:
                cbar_min = np.around(Z.min(),decimals=1)        # colorbar min value - avoid going out of range
                cbar_max = 0                                    # colorbar max value
            plt.text(0.81,1.6,str(cbar_max),size=15)         # easier to show cbar labels as text
            plt.text(0.73,-14.2,str(cbar_min),size=15)
        cbar.set_ticks([]) # comment if you want ticks
        cbar.ax.set_yticklabels('') # comment if you want tick labels
        cbar.ax.set_ylabel(cbar_label,fontsize='xx-large')

        return plt
        

    def add_reservoirs(self,reservoirs,elements,size=1,label_size=1,color=[],edgecolor='k',
                       label_color='k',linewidths=3,**kwargs):
        """
        Add reservoirs as points on the stability diagram.

        Parameters
        ----------
        reservoirs : Reservoirs
            Reservoirs object.
        elements : list
            List with strings of the elements to be used as free variables.
        size : float
            Float multiplier for points size. Default is 1, which would yield a default size of 450*self.size
        label_size : float
            Float multiplier for labels size. Default is 1, which would yield a default size of 30*self.size
        color : str
            Color of filling of points
        edgecolor : str 
            Color of point edge
        label_color : str 
            Color of labels
        linewidths : 3
            line width of point edge
        kwargs: dict
            kwargs to pass to matplotlib `scatter` function.

        Returns
        -------
        plt : matplotlib
            Matplotlib object.

        """         
        points = {}
        for r,mu in reservoirs.items():
            points[r] = [mu[el] for el in elements]
        
        return self.add_points(points,size,label_size,color,edgecolor,label_color,linewidths,**kwargs)
        
    
    def constant_chempot_line(self, mu, comp, variable_element, chempots_ref):
        """
        Function that expresses line of constant chemical potential of a given composition. Only works for 3-component PD.

        Parameters
        ----------
        mu : float
            Indipendent variable of chemical potential.
        comp : str or Composition
            Composition of the phase.
        variable_element : Pymatgen Element object
            Element chosen as indipendent variable.
        chempots_ref : dict
            Dictionary with fixed chemical potentials (values relative to reference phase). The format is {Element:chempot}.
        
        Returns
        -------
        chempot : float
            Chemical potential in eV.

        """
        comp = _get_composition_object(comp)
        chempots_ref[variable_element] = mu
        return self.chempots_analysis.calculate_single_chempot(comp,chempots_ref)
    
    
    def get_plot(self,elements,figsize=None):
        """
        Method to get stability diagram with 'get_chempot_range_map_plot' method in pymatgen.

        Parameters
        ----------
        elements : list
            List with strings of the elements to be used as free variables.
        figsize : tuple
            New size in inches.

        Returns
        -------
        plt : matplotlib
            Matplotlib object.

        """
        pdhandler = PDHandler(self.phase_diagram)
        return pdhandler.get_stability_diagram(elements=elements,figsize=figsize)
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 15:24:29 2021

@author: villa
"""
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from .defects import Defect, format_legend_with_charge_number, get_defect_from_string



def plot_formation_energies(entries,
                            chemical_potentials,
                            vbm,
                            band_gap,
                            temperature=0,
                            xlim=None,
                            ylim=None,
                            title=None,
                            fermi_level=None,
                            grid=True,
                            figsize=(6,6),
                            fontsize=12,
                            show_legend=True,
                            format_legend=True,
                            get_subplot=False,
                            subplot_settings=None,
                            **eform_kwargs):
    """
    Produce defect Formation energy vs Fermi energy plot.

    Parameters
    ----------
    entries : list 
        List of entries to calculate.
    chemical_potentials : dict
        Dictionary with chemical potentials of the elements {'element':chempot}.
    vbm : float
        Valence band maximum in eV.
    band_gap : float
        Band gap of bulk material in eV.
    temperature : float
        Temperature in K. If no custom formation energy is provided, this arg has no effect. 
    xlim : tuple
        Tuple giving the range of the x (fermi energy) axis.
    ylim : tuple
        Tuple giving the range for the formation energy axis.
    title : str
        Title of the figure.
    fermi_level : float
        Plot Fermi energy position with a vertical line.
    grid : bool
        Show grid.
    figsize : float or tuple
        Figure size.
    fontsize : float
        Font size.
    show_legend  : bool
        Show legend.
    format_legend : bool
        Get latex-like legend based on the name of defect entries.
    get_subplot : bool
        Get subplot.
    subplot_settings:
        List with integers for subplot setting on matplotlib (plt.subplot(nrows,ncolumns,index)). 
    eform_kwargs : dict
        Kwargs to pass to `entry.formation_energy`.

    Returns
    -------
    matplotlib object

    """
    from .analysis import DefectsAnalysis
    import matplotlib.pyplot as plt
    
    matplotlib.rcParams.update({'font.size': fontsize}) 
    formation_energies = DefectsAnalysis(
                        entries=entries,
                        band_gap=band_gap,
                        vbm=vbm).formation_energies(chemical_potentials=chemical_potentials,
                                                                                    fermi_level=0,
                                                                                    temperature=temperature,
                                                                                    entries=entries,
                                                                                    **eform_kwargs)
    if xlim == None:
        xlim = (-0.5,band_gap+0.5)        
    npoints = 200
    step = abs(xlim[1]+0.1-xlim[0])/npoints
    x = np.arange(xlim[0],xlim[1]+0.1,step)
    
    if get_subplot:
        if subplot_settings[2] == 1:
            plt.figure(figsize=figsize)               
        plt.subplot(subplot_settings[0],subplot_settings[1],subplot_settings[2])
    else:
        plt.figure(figsize=figsize)
        
    for name in formation_energies:
        energy = np.zeros(len(x))
        emin = np.zeros(len(x))
        x_star = []
        y_star = []
        q_previous = None
        for i in range(0,npoints):
            emin[i] = 1e40
            for d in formation_energies[name]:
                q = d[0]
                e0 = d[1]
                energy[i] = e0 + q*x[i]
                # finding most stable charge state
                if energy[i] < emin[i]:
                    emin[i] = energy[i]
                    q_stable = q
            # getting data to plot transition levels        
            if q_stable != q_previous:
                if q_previous != None:
                    x_star.append(x[i])
                    y_star.append(emin[i])
                q_previous = q_stable       

        if format_legend:
            label_txt = get_defect_from_string(name).symbol
        else:
            label_txt = name            

        plt.plot(x,emin,label=label_txt,linewidth=3)
        plt.scatter(x_star,y_star,s=120,marker='*')
                    
    plt.axvline(x=0.0, linestyle='-', color='k', linewidth=2)  # black dashed lines for gap edges
    plt.axvline(x=band_gap, linestyle='-', color='k',
                linewidth=2)        
    if fermi_level:
        plt.axvline(x=fermi_level, linestyle='dashed', color='k', linewidth=1.5, label='$\mu _{e}$')                
    # shaded areas
    plt.axvspan(xlim[0], 0, facecolor='k', alpha=0.2)
    plt.axvspan(band_gap, xlim[1]+0.1, facecolor='k', alpha=0.2)
    plt.hlines(0,xlim[0],xlim[1]+0.1,colors='k',linestyles='dashed',alpha=0.5)
    plt.xlim(xlim)
    if ylim: 
        plt.ylim(ylim) 
    plt.xlabel('Fermi level (eV)')
    plt.ylabel('Formation energy (eV)')
    if title:
        plt.title(title)
    if show_legend:    
        plt.legend()
    if grid:
        plt.grid()

    return plt


def plot_binding_energies(entries,
                          vbm,
                          band_gap,
                          temperature=0,
                          names=None,
                          xlim=None,
                          ylim=None,
                          figsize=(6,6),
                          fontsize=18,
                          format_legend=True,
                          **eform_kwargs):
    """
    Plot binding energies for complex of defects as a function of the fermi level

    Parameters
    ----------
    entries : list 
        List of entries to calculate.
    vbm : float
        Valence band maximum in eV.
    band_gap : float
        Band gap of bulk material in eV.
    temperature : float
        Temperature in K. If no custom formation energy is provided, this arg has no effect.
    names : list
        List of strings with names of DefectEntry. If None, all defect complexes are plotted.
    xlim : tuple
        Tuple giving the range of the x (fermi energy) axis
    ylim : tuple
        Tuple giving the range for the formation energy axis
    figsize : tuple
        Figure size.
    fontsize : float
        Font size.
    format_legend : bool
        Bool for getting latex-like legend based on the name of defect entries.
    eform_kwargs : dict
        Kwargs to pass to `entry.formation_energy`.

    Returns
    -------
    matplotlib object

    """         
    from .analysis import DefectsAnalysis
    
    da = DefectsAnalysis(entries=entries,band_gap=band_gap,vbm=vbm)
    plt.figure(figsize=figsize)
    matplotlib.rcParams.update({'font.size': fontsize}) 
    if xlim==None:
        xlim = (-0.5,da.band_gap+0.5)
    # building array for x values (fermi level)    
    ef = np.arange(xlim[0],xlim[1]+0.1,(xlim[1]-xlim[0])/200)        
    binding_energy = np.zeros(len(ef))        
    if not names:
        names = []
        for e in da.entries:
            if e.defect_type == 'DefectComplex':
                if e.name not in names:
                    names.append(e.name)   
    if not names:
        raise ValueError('No DefectComplex entries found')  

    # getting binding energy at different fermi levels for every name in list
    for name in names:
        label = da.select_entries(names=[name])[0].symbol if format_legend else name
        for i in range(0,len(ef)):
            binding_energy[i] = da.binding_energy(
                                        name=name,
                                        fermi_level=ef[i],
                                        temperature=temperature,
                                        **eform_kwargs)
                        
        plt.plot(ef,binding_energy, linewidth=2.5*(figsize[1]/figsize[0]),label=label)
        
    plt.axvline(x=0.0, linestyle='-', color='k', linewidth=2)  # black lines for gap edges
    plt.axvline(x=da.band_gap, linestyle='-', color='k',
                linewidth=2)        
    # shaded areas
    plt.axvspan(xlim[0], 0, facecolor='k', alpha=0.2)
    plt.axvspan(da.band_gap, xlim[1], facecolor='k', alpha=0.2)
    plt.hlines(0,xlim[0],xlim[1],colors='k',linestyles='dashed',alpha=0.5)
    plt.legend()
    plt.xlim(xlim)
    if ylim: 
        plt.ylim(ylim) 
    plt.xlabel('Fermi level (eV)')
    plt.ylabel('Binding energy (eV)')
    
    return plt


def plot_charge_transition_levels(entries,
                                  vbm,
                                  band_gap,
                                  temperature=0,
                                  ylim=None,
                                  figsize=(6,6),
                                  fontsize=16,
                                  fermi_level=None,
                                  format_legend=True,
                                  get_integers=True,
                                  **eform_kwargs):
    """
    Plotter for the charge transition levels.

    Parameters
    ----------
    entries : list
        List of entries to calculate.
    vbm : float
        Valence band maximum in eV.
    band_gap : float
        Band gap of bulk material in eV.
    temperature : float
        Temperature in K. If no custom formation energy is provided, this arg has no effect.
    ylim : tuple
        y-axis limits.
    figsize : tuple
        Figure size.
    fontsize : float
        Font size.
    fermi_level : float
        Plot Fermi energy position.
    format_legend : bool
        Bool for getting latex-like legend based on the name of defect entries.
    get_integers : bool
        Get charge transition levels as integers.
    eform_kwargs : dict
        Kwargs to pass to `entry.formation_energy`.

    Returns
    -------
    matplotlib object

    """        
    from .analysis import DefectsAnalysis
    
    da = DefectsAnalysis(entries=entries,band_gap=band_gap,vbm=vbm)
    plt.figure(figsize=figsize)         
    if ylim == None:
        ylim = (-0.5,da.band_gap +0.5)        
    charge_transition_levels = da.charge_transition_levels(
                                                temperature=temperature,
                                                entries=entries,
                                                get_integers=get_integers,
                                                **eform_kwargs)
    
    number_defects = len(charge_transition_levels)   
    x_max = 10
    interval = x_max/(number_defects + 1)
    x = np.arange(0,x_max,interval)
    # position of x labels
    x_ticks_positions = []
    for i in range(0,len(x)-1):
        x_ticks_positions.append((x[i+1]-x[i])/2 + x[i])            
    x_ticks_labels = []
    for name in charge_transition_levels:
        x_ticks_labels.append(name)        
    # draw vertical lines to separte defect types
    for i in x:
        plt.axvline(x=i, linestyle='-', color='k', linewidth=1.2, alpha=1, zorder=1)
    xlim = (x[0],x[-1])        
    #VBM and CBM shaded
    plt.axhspan(ylim[0], 0, facecolor='grey', alpha=0.9, zorder=2)
    plt.axhspan(da.band_gap,ylim[1], facecolor = 'grey', alpha=0.9, zorder=2)                
    # plot CTL
    for i in range(0,len(x_ticks_labels)):
        name = x_ticks_labels[i]
        for ctl in charge_transition_levels[name]:
            energy = ctl[2]
            plt.hlines(energy,x[i],x[i+1],colors='k',linewidth=2.25, zorder=3)
            charge1 = '+' + str(ctl[1]) if ctl[1] > 0 else str(ctl[1])
            charge2 = '+' + str(ctl[0]) if ctl[0] > 0 else str(ctl[0])
            label_charge = '(' + charge2 + '/' + charge1 + ')'
            font_space = abs(ylim[1]-ylim[0]) / 100
            if energy < ylim[1] and energy > ylim[0]:
                plt.text(x[i]+(interval/2)*2/number_defects ,energy+font_space,label_charge,fontsize=fontsize)        
    # format latex-like legend
    if format_legend:    
         for name in x_ticks_labels:            
            x_ticks_labels[x_ticks_labels.index(name)] = get_defect_from_string(name).symbol               
    if fermi_level:
        plt.axhline(y=fermi_level, linestyle='dashed', color='k', linewidth=1.5, label='$\mu _{e}$')   
    
    plt.text(x[-1]+interval/8,-0.3,'VB',fontsize=25*(fontsize/16))
    plt.text(x[-1]+interval/8,da.band_gap+0.2,'CB',fontsize=25*(fontsize/16))
    plt.xticks(ticks=x_ticks_positions,labels=x_ticks_labels,fontsize = (25-number_defects)*(fontsize/16))
    plt.tick_params(axis='x',length=0,width=0)
    plt.yticks(fontsize=16*(fontsize/16))
    plt.xlim(xlim)  
    plt.ylim(ylim)
    plt.ylabel('Energy(eV)',fontsize=20*(fontsize/16))  
    
    return plt    


def plot_pO2_vs_concentrations(
                            thermodata,
                            output='total',
                            figsize=(8,8),
                            fontsize=22,
                            xlim=(1e-20,1e10),
                            ylim=None,
                            colors=None,
                            **kwargs):
    """
    Plot defect and carrier concentrations in a range of oxygen partial pressure.

    Parameters
    ----------
    thermodata: ThermoData
        ThermoData object that contains the thermodynamic data:

        partial_pressures : (list)
            list with partial pressure values.
        defect_concentrations : (list or dict)
            Defect concentrations in the same format as the output of DefectsAnalysis. 
        carrier_concentrations : (list)
            List of tuples with intrinsic carriers concentrations (holes,electrons).

    output : str
        Type of output for defect concentrations:

        - "all": The output is the concentration of every defect entry.
        - "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
        - "total": The output is the sum of the concentration in every charge for each specie.

    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams.
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis. 
    colors : list
        List of colors to use for plotting with matplotlib. If None the defaults are used.
    kwargs : dict
        Kwargs to pass to `DefectConcentrations.filter_concentrations(**kwargs)`.
        If provided, only the filtered concentrations will be plotted. If output
        is set to "total", only the filtered concentrations will be used to 
        compute the total concentration.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    td = thermodata
    p,dc,cc = td.partial_pressures,td.defect_concentrations,td.carrier_concentrations
    xlabel = 'Oxygen partial pressure (atm)'
    plt = plot_x_vs_concentrations(
                                x=p,
                                xlabel=xlabel,
                                defect_concentrations=dc,
                                carrier_concentrations=cc,
                                output=output,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim,
                                colors=colors,
                                **kwargs)
    return plt


def plot_pO2_vs_conductivity(
                            partial_pressures,
                            conductivities,
                            new_figure=True,
                            label=None,
                            figsize=(8,8),
                            fontsize=22,
                            xlim=(1e-20,1e10),
                            ylim=None):
    """
    Plot conductivity as a function of the oxygen partial pressure.

    Parameters
    ----------
    partial_pressures : list
        list with partial pressure values.
    conductivities : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and conductivity list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    new_figure : bool
        Initialize a new matplotlib figure.
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    xlabel = 'Oxygen partial pressure (atm)'
    plt = plot_x_vs_conductivity(
                                x=partial_pressures,
                                xlabel=xlabel,
                                conductivities=conductivities,
                                new_figure=new_figure,
                                label=label,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim)
    return plt    



def plot_pO2_vs_fermi_level(
                            partial_pressures,
                            fermi_levels,
                            band_gap,
                            new_figure=True,
                            label=None,
                            figsize=(8,8),
                            fontsize=22,
                            xlim=(1e-20,1e10),
                            ylim=None,
                            colors=None):
    """
    Plot Fermi level as a function of the oxygen partial pressure.

    Parameters
    ----------
    partial_pressures : list
        list with partial pressure values.
    fermi_levels : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and fermi level list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    band_gap : float
        Band gap of the bulk material.
    new_figure : bool
        Initialize a new matplotlib figure.
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.
    colors : list
        List with colors for Fermi level data.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    xlabel = 'Oxygen partial pressure (atm)'
    plt = plot_x_vs_fermi_level(
                                x=partial_pressures,
                                xlabel=xlabel,
                                fermi_levels=fermi_levels,
                                band_gap=band_gap,
                                new_figure=new_figure,
                                label=label,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim,
                                colors=colors)
    return plt


def plot_variable_species_vs_concentrations(
                                        thermodata,
                                        output='total',
                                        figsize=(8,8),
                                        fontsize=22,
                                        xlim=(1e-20,1e10),
                                        ylim=None,
                                        colors=None,
                                        **kwargs):
    """
    Plot defect and carrier concentrations in a range of oxygen partial pressure.

    Parameters
    ----------
    thermodata: ThermoData
        ThermoData object that contains the thermodynamic data:

        partial_pressures : (list)
            list with partial pressure values.
        defect_concentrations : (list or dict)
            Defect concentrations in the same format as the output of DefectsAnalysis. 
        carrier_concentrations : (list)
            List of tuples with intrinsic carriers concentrations (holes,electrons).

    output : str
        Type of output for defect concentrations:

        - "all": The output is the concentration of every defect entry.
        - "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
        - "total": The output is the sum of the concentration in every charge for each specie.

    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams.
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.
    colors : list
        List of colors to use for plotting with matplotlib. If None the defaults are used.
    kwargs : dict
        Kwargs to pass to `DefectConcentrations.filter_concentrations(**kwargs)`.
        If provided, only the filtered concentrations will be plotted. If output
        is set to "total", only the filtered concentrations will be used to 
        compute the total concentration.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.
    """
    td = thermodata
    c,dc,cc = td.variable_concentrations,td.defect_concentrations,td.carrier_concentrations
    dname = _get_variable_defect_specie_label(td.variable_defect_specie)
    xlabel = '[%s] (cm$^{-3})$' %dname
    
    plt = plot_x_vs_concentrations(
                                x=c,
                                xlabel=xlabel,
                                defect_concentrations=dc,
                                carrier_concentrations=cc,
                                output=output,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim,
                                colors=colors,
                                **kwargs)
    return plt


def plot_variable_species_vs_conductivity(
                                        xlabel,
                                        variable_concentrations,
                                        conductivities,
                                        new_figure=True,
                                        label=None,
                                        figsize=(8,8),
                                        fontsize=22,
                                        xlim=(1e-20,1e10),
                                        ylim=None):
    """
    Plot conductivity as a function of the oxygen partial pressure.

    Parameters
    ----------
    xlabel : str
        Label for concentration axis (cm^-3 is added).
    variable_concentrations : list
        List of concentrations of variable species. 
    conductivities : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and conductivity list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    new_figure : bool
        Initialize a new matplotlib figure.
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.
    """
    xlabel = '%s (cm$^{-3})$' %xlabel
    plt = plot_x_vs_conductivity(
                                x=variable_concentrations,
                                xlabel=xlabel,
                                conductivities=conductivities,
                                new_figure=new_figure,
                                label=label,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim)
    return plt
    

def plot_variable_species_vs_fermi_level(
                                        xlabel,
                                        variable_concentrations,
                                        fermi_levels,
                                        band_gap,
                                        new_figure=True,
                                        label=None,
                                        figsize=(8,8),
                                        fontsize=22,
                                        xlim=(1e-20,1e10),
                                        ylim=None,
                                        colors=None):
    """
    Plot Fermi level as a function of the oxygen partial pressure.

    Parameters
    ----------
    xlabel : str
        Label for concentration axis (cm^-3 is added).
    variable_concentrations : list
        List of concentrations of variable species. 
    fermi_levels : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and fermi level list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    band_gap : float
        Band gap of the bulk material.
    new_figure : bool
        Initialize a new matplotlib figure.
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.
    colors : list
        List with colors for Fermi level data.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    xlabel = '%s (cm$^{-3})$' %xlabel
    plt = plot_x_vs_fermi_level(
                                x=variable_concentrations,
                                xlabel=xlabel,
                                fermi_levels=fermi_levels,
                                band_gap=band_gap,
                                new_figure=new_figure,
                                label=label,
                                figsize=figsize,
                                fontsize=fontsize,
                                xlim=xlim,
                                ylim=ylim,
                                colors=colors)
    return plt



def plot_x_vs_concentrations(
                            x,
                            xlabel,
                            defect_concentrations,
                            carrier_concentrations,
                            output='total',
                            figsize=(8,8),
                            fontsize=22,
                            xlim=(1e-20,1e10),
                            ylim=None,
                            colors=None,
                            **kwargs):
    """
    Plot defect concentrations as a function of generic data on the x-axis.        
    
    Parameters
    ----------
    x : list
        List with data on x-axis.
    xlabel : str
        Label for x-axis.
    defect_concentrations : list
        List of DefectConcentrations objects.
    carrier_concentrations : list
        List of tuples with carrier concentrations (holes,electrons).
    output : str
        Type of output for defect concentrations:

        - "all": The output is the concentration of every defect entry.
        - "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
        - "total": The output is the sum of the concentration in every charge for each specie.
        
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis. 
    ylim : tuple
        Range of y-axis.
    colors : list
        List of colors to use for plotting with matplotlib. If None the defaults are used.
    kwargs : dict
        Kwargs to pass to `DefectConcentrations.filter_concentrations(**kwargs)`.
        If provided, only the filtered concentrations will be plotted. If output
        is set to "total", only the filtered concentrations will be used to 
        compute the total concentration.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

        """        
    matplotlib.rcParams.update({'font.size': 22})
    if output == 'all' or output == 'stable':
        plt = _plot_conc(
                        x=x,
                        defect_concentrations=defect_concentrations,
                        carrier_concentrations=carrier_concentrations,
                        output=output,figsize=figsize,colors=colors,**kwargs)
        
    elif output == 'total':
        plt = _plot_conc_total(
                            x=x,
                            defect_concentrations=defect_concentrations,
                            carrier_concentrations=carrier_concentrations,
                            figsize=figsize,colors=colors,**kwargs)
    else:
        raise ValueError('The options for plot output are "all", "stable" or "total".')
        
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)

    plt.xlabel(xlabel)
    plt.ylabel('Concentrations (cm$^{-3})$')
    plt.legend()
    plt.grid()
    
    return plt   

    

def plot_x_vs_conductivity(
                        x,
                        xlabel,
                        conductivities,
                        new_figure=True,
                        label=None,
                        figsize=(8,8),
                        fontsize=22,
                        xlim=(1e-20,1e10),
                        ylim=None):
    """
    Plot conductivity as a function of the oxygen partial pressure.

    Parameters
    ----------
    x : list
        List with data on x-axis.
    xlabel : str
        Label for x-axis.
    conductivities : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and conductivity list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    new_figure : bool
        Initialize a new matplotlib figure. 
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis.
    ylim : tuple
        Range of y-axis.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    matplotlib.rcParams.update({'font.size': fontsize})
    if new_figure:
        plt.figure(figsize=figsize)
    if isinstance(conductivities,dict):
        for name,sigma in conductivities.items():
            plt.plot(x,sigma,linewidth=4,marker='s',label=name)
    else:
        sigma = conductivities
        plt.plot(x,sigma,linewidth=4,marker='s',label=label)
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)
    plt.xlabel(xlabel)
    plt.ylabel('Conductivity (S/m)')
    plt.legend()
    if new_figure:
        plt.grid()
    
    return plt


def plot_x_vs_fermi_level(
                        x,
                        xlabel,
                        fermi_levels,
                        band_gap,
                        new_figure=True,
                        label=None,
                        figsize=(8,8),
                        fontsize=22,
                        xlim=(1e-20,1e10),
                        ylim=None,
                        colors=None):
    """
    Parameters
    ----------
    x : list
        List with data on x-axis.
    xlabel : str
        Label for x-axis.
    fermi_levels : dict or list
        If is a dict multiples lines will be plotted, with labels as keys and fermi level list
        as values. If is a list only one line is plotted with label taken from the "label" argument.
    band_gap : float
        Band gap of the bulk material.
    new_figure : bool
        Initialize a new matplotlib figure.
    label : str
        Label for the data.
    figsize : tuple
        Size of the matplotlib figure.
    fontsize : float
        Size of font for matplotlib rcParams. 
    xlim : tuple
        Range of x-axis. 
    ylim : tuple
        Range of y-axis.
    colors : list,
        List with colors for Fermi level data.

    Returns
    -------
    plt : matplotlib
        Matplotlib object.

    """
    ylim = ylim if ylim else (-0.5, band_gap+0.5)
    matplotlib.rcParams.update({'font.size': fontsize})
    if new_figure:
        plt.figure(figsize=figsize)
    if isinstance(fermi_levels,dict):
        for name,mue in fermi_levels.items():
            clr = colors[list(fermi_levels.keys()).index(name)] if colors else None
            plt.plot(x,mue,linewidth=4,label=name,color=clr)
    else:
        mue = fermi_levels
        clr = colors[0] if colors else None
        plt.plot(x,mue,linewidth=4,label=label,color=clr)
    plt.xscale('log')
    plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)
    plt.hlines(0, xlim[0], xlim[1],color='k')
    plt.text(xlim[1]+(xlim[1]-xlim[0]),-0.05,'VB')
    plt.text(xlim[1]+(xlim[1]-xlim[0]),band_gap-0.05,'CB')
    plt.hlines(band_gap, xlim[0], xlim[1],color='k')
    plt.axhspan(ylim[0], 0, alpha=0.2,color='k')
    plt.axhspan(band_gap, ylim[1], alpha=0.2,color='k')

    plt.xlabel(xlabel)
    plt.ylabel('Electron chemical potential (eV)')
    if type(fermi_levels) == dict:
        plt.legend()
    if new_figure:
        plt.grid()
    
    return plt


def _get_variable_defect_specie_label(variable_defect_specie):
    try:
        defect = get_defect_from_string(variable_defect_specie)
        return defect.symbol
    except:
        return variable_defect_specie



def _plot_conc(
            x,
            defect_concentrations,
            carrier_concentrations,
            output,
            figsize,
            colors,
            **kwargs):
    
    plt.figure(figsize=figsize)
    dc = defect_concentrations if output != 'stable' else [c.stable for c in defect_concentrations] 
    if kwargs:
        dc = [c.filter_concentrations(**kwargs) for c in dc] #filter concentrations based on kwargs
    h = [cr[0] for cr in carrier_concentrations] 
    n = [cr[1] for cr in carrier_concentrations]
    previous_charge = None
    for i in range(0,len(dc[0])):
            conc = [c[i].conc for c in dc]
            charges = [c[i].charge for c in dc]
            label_txt = get_defect_from_string(dc[0][i].name).symbol
            if output == 'all':
                label_txt = format_legend_with_charge_number(label_txt,dc[0][i].charge)
            elif output == 'stable':
                for q in charges:
                    if q != previous_charge:
                        previous_charge = q
                        label_charge = '+' + str(int(q)) if q > 0 else str(int(q))
                        index = charges.index(q)
                        plt.text(x[index],conc[index],label_charge,clip_on=True)
            color = colors[i] if colors else None
            plt.plot(x,conc,label=label_txt,linewidth=4,color=color)
    plt.plot(x,h,label='$n_{h}$',linestyle='--',color='r',linewidth=4)
    plt.plot(x,n,label='$n_{e}$',linestyle='--',color='b',linewidth=4)
    return plt


def _plot_conc_total(
                    x,
                    defect_concentrations,
                    carrier_concentrations,
                    figsize,
                    colors,
                    **kwargs):
    
    dc = defect_concentrations
    if kwargs:
        dc = [c.filter_concentrations(**kwargs) for c in dc]
    plt.figure(figsize=figsize)
    h = [cr[0] for cr in carrier_concentrations] 
    n = [cr[1] for cr in carrier_concentrations]
    for name in dc[0].names:
        conc = [c.total[name] for c in dc]
        label_txt = get_defect_from_string(name).symbol
        color = colors[dc[0].names.index(name)] if colors else None
        plt.plot(x,conc,label=label_txt,linewidth=4,color=color)
    plt.plot(x,h,label='$n_{h}$',linestyle='--',color='r',linewidth=4)
    plt.plot(x,n,label='$n_{e}$',linestyle='--',color='b',linewidth=4)
    return plt


def _get_unstable_bool(defect_concentrations):
    slist = []
    for dc in defect_concentrations:
        unstable = False
        for d in dc:
            if 'stable' in vars(d).keys() and d.stable == False:
                unstable = True
        slist.append(unstable)
    
    return slist
        
    

class DefectConcentrationsPlotter:

    def __init__(self,concentrations,format_names=True):
        """
        Class to visualize concentrations dictionaries with pd.DataFrame and 
        pd.Series (for total concentrations).

        Parameters
        ----------
        concentrations : list or dict
            Concentrations can be in different formats: list of dict for "all" and 
            "stable_charges" and dict for "total".
        format_names : bool
            Format names with latex symbols.
        """
        # to be fixed
        conc_dict = []
        for c in concentrations:
            d = {'charge':c.charge,'conc':c.conc}

            if format_names:
                name = get_defect_from_string(c.name).symbol
            else:
                name = c.name
            d['name'] = format_legend_with_charge_number(name,c.charge)
            conc_dict.append(d)
        
        conc_total_dict = {}
        for dn,conc in concentrations.total.items():
            if format_names:
                name = get_defect_from_string(dn).symbol
            else:
                name = dn
            conc_total_dict[name] = conc

        self.conc = conc_dict
        self.conc_total = conc_total_dict
        self.format = format_names
        self.df = pd.DataFrame(self.conc)
        self.series = pd.Series(self.conc_total,name='Total Concentrations')
    

    def __print__(self):
        return self.df.__print__()
    
    def __repr__(self):
        return self.df.__repr__()

    def copy(self):
        return self.df.copy()
    
    
    def plot_bar(self,conc_range=(1e13,1e40),ylim=None,total=True,ylabel_fontsize=15,**kwargs):
        """
        Bar plot of concentrations with `pd.DataFrame.plot`

        Parameters
        ----------
        conc_range : tuple
            Range of concentrations to include in df.
        ylim : tuple
            Limit of y-axis in plot. If None conc_range is used.
        total : bool
            plot total concentrations.
        ylabel_fontsize : int
            Size of the label for y-axis.
        **kwargs : dict
            Kwargs to pass to `df.plot`.

        Returns
        -------
        plt : 
            Matplotlib object.

        """
        if conc_range:
            if not ylim:
                ylim = conc_range
            series = self.limit_conc_range(conc_range,reset_df=False)[1]
            df = self.limit_conc_range(conc_range,reset_df=False)[0]
        else:
            series = self.series
            df = self.df
        if not total:
            ax = df.plot(x='name',y='conc',kind='bar',ylim=ylim,logy=True,grid=True,
                          xlabel='Name , Charge',legend=None,**kwargs)
        else:
            ax = series.plot(kind='bar',logy=True,ylim=ylim,ylabel='Concentrations(cm$^{-3}$)',grid=True,**kwargs)
        ax.set_ylabel('Concentrations(cm$^{-3}$)',fontdict={'fontsize':ylabel_fontsize})
        return ax


    def limit_conc_range(self,conc_range=(1e10,1e40),reset_df=False):
        """
        Limit df to a concentration range.

        Parameters
        ----------
        conc_range : tuple
            Range of concentrations to include in df.
        reset_df : bool
            Reset df attribute or return a new df.

        Returns
        -------
        df, series : tuple 
            DataFrame object, Series object.
            
        """
        if reset_df:
            self.df = self.df[self.df.conc.between(conc_range[0],conc_range[1])]
            self.series = self.series.between(conc_range[0],conc_range[1])
            return
        else:
            df = self.df.copy()
            series = self.series.copy()
            df = df[df.conc.between(conc_range[0],conc_range[1])]
            series = series[series.between(conc_range[0],conc_range[1])]
            return df , series    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  2 16:33:50 2025

@author: lorenzo
"""

import warnings
import json
import os.path as op
from pandas import DataFrame
from monty.json import MSONable, MontyEncoder
import copy

from pymatgen.analysis.phase_diagram import PhaseDiagram

from .core import Chempots
from .phase_diagram import PDHandler
from ..tools.utils import format_composition


class Reservoirs(MSONable):
    
    def __init__(self,res_dict,phase_diagram=None,mu_refs=None,are_chempots_delta=False):
        """
        Class to handle dictionaries of chemical potentials. Works almost completely like a python dictionary.

        Parameters
        ----------
        res_dict : dict
            Dictionary with reservoir names as key and dictionaries of chemical potentials as values.
        phase_diagram : PhaseDiagram
            PhaseDiagram object, useful to convert absolute chempots in referenced chempots.
        mu_refs : dict
            Dictionary with chemical potentials of reference elements ({Element:chempot}).
            If the PhaseDiagram is provided mu_refs is taken from the mu_refs attribute. 
        are_chempots_delta : bool
            Set this variable to True if chempots in dictionary are referenced values.

        """
        self.res_dict = res_dict
        self.pd = phase_diagram 
        if mu_refs:
            self.mu_refs = mu_refs
        elif self.pd:
            self.mu_refs = PDHandler(self.pd).get_chempots_reference()
        else:
            self.mu_refs = mu_refs
            warnings.warn('Neither PhaseDiagram or reference chempots have been provided, conversions btw ref and abs value will not be possible',UserWarning)
        self._are_chempots_delta = are_chempots_delta


    def __str__(self):
        df = self.get_dataframe()
        return df.__str__()
    
    def __repr__(self):
        return self.__str__()
    
    def _repr_html_(self):
        return self.get_dataframe()._repr_html_()

    def __len__(self):
        return len(self.res_dict)

    def __iter__(self):
        return self.res_dict.keys().__iter__()

    def __getitem__(self,reskey):
        return self.res_dict[reskey]

    def __setitem__(self,reskey,chempots):
        self.res_dict[reskey] = chempots
        return
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.res_dict == other
        elif isinstance(other, Reservoirs):
            return self.res_dict == other.res_dict
        else:
            return False

    def keys(self):
        return self.res_dict.keys()

    def values(self):
        return self.res_dict.values()
    
    def items(self):
        return self.res_dict.items()

    def copy(self):
        return Reservoirs(copy.deepcopy(self.res_dict),phase_diagram=self.pd,
                          are_chempots_delta=self.are_chempots_delta,mu_refs=self.mu_refs)
    
    def update(self, other):
        if isinstance(other, dict):
            for key, value in other.items():
                self.res_dict[key] = value
        else:
            for key, value in other:
                self.res_dict[key] = value


    @property
    def are_chempots_delta(self):
        return self._are_chempots_delta

    def as_dict(self):
        """
        Json-serializable dict representation of a Reservoirs object. 
        
        Returns
        -------
        d : dict
            Json-serializable dict of a Reservoirs object.

        """
        d = {}
        d['@module'] = self.__class__.__module__
        d['@class'] = self.__class__.__name__
        d['res_dict'] = {r:mu.as_dict() for r,mu in self.res_dict.items()} 
        d['phase_diagram'] = self.pd.as_dict() if self.pd else None
        d['mu_refs'] = self.mu_refs.as_dict() if self.mu_refs else None
        d['are_chempots_delta'] = self.are_chempots_delta
        return d


    def to_json(self,path,cls=MontyEncoder):
        """
        Save Reservoirs object as json string or file

        Parameters
        ----------
        path : str
            Path to the destination file. If None a string is exported.
        cls : cls
            Encoder class for json.dump. The default is MontyEncoder.

        Returns
        -------
        d : str
            If path is not set a string is returned.

        """
        d = self.as_dict()
        if path:
            with open(path,'w') as file:
                json.dump(d,file,cls=cls)
            return
        else:
            return d.__str__()  


    @classmethod
    def from_dict(cls,d):
        """
        Constructor of Reservoirs object from dictionary representation.
        
        Parameters
        ----------
        d : dict
            Json-serializable dict of a Reservoirs object.
        
        Returns
        -------
        reservoirs : Reservoirs
            Reservoirs object.

        """
        res_dict = {}
        for res,chempots in d['res_dict'].items():
            res_dict[res] = Chempots.from_dict(chempots)
        phase_diagram = PhaseDiagram.from_dict(d['phase_diagram']) if d['phase_diagram'] else None
        mu_refs = Chempots.from_dict(d['mu_refs']) if d['mu_refs'] else None
        are_chempots_delta = d['are_chempots_delta']
            
        return cls(res_dict,phase_diagram,mu_refs,are_chempots_delta)


    @staticmethod
    def from_json(path_or_string):
        """
        Build Reservoirs object from json file or string.

        Parameters
        ----------
        path_or_string : str
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        Reservoir object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return Reservoirs.from_dict(d)


    def filter_reservoirs(self,inplace=False,elements=None):
        """
        Get new Reservoir object filtering the chempots dictionary.

        Parameters
        ----------
        inplace : bool
            Apply changes to current Reservoirs object.
        elements : list
            List of element symbols.

        Returns
        -------
        res : Reservoirs
            Reservoirs object.

        """
        res = self.copy()
        mu_refs = self.mu_refs.copy()
        filtered_dict = res.res_dict
        if elements:
            d = filtered_dict.copy()
            for r in list(d):
                for el in list(d[r]):
                    if el not in elements:
                        del filtered_dict[r][el]
                        if el in mu_refs.keys():
                            del mu_refs[el]
        if inplace:
            self.res_dict = filtered_dict
            self.mu_refs = mu_refs
            return
        else:
            res.mu_refs = mu_refs
            return res


    def get_absolute_res_dict(self):
        """ 
        Return values of chempots from referenced to absolute 
        """
        if self.are_chempots_delta:
            return {r:mu.get_absolute(self.mu_refs) for r,mu in self.res_dict.items()}
        else:
            raise ValueError('Chemical potential values are already absolute')
                

    def get_referenced_res_dict(self):
        """ 
        Return values of chempots from absolute to referenced 
        """
        if self.are_chempots_delta:
            raise ValueError('Chemical potential values are already with respect to reference')
        else:
            return {r:mu.get_referenced(self.mu_refs) for r,mu in self.res_dict.items()}

    
    def get_dataframe(self,format_symbols=False,format_compositions=False,all_math=False,ndecimals=None):
        """
        Get DataFrame object of the dictionary of reservoirs.

        Parameters
        ----------
        format_symbols : bool
            Format labels of element chempots in latex math format.
        format_compositions : bool
            Get Latex format of compositions.
        all_math : bool
            Get all characters in composition written in latex's math format.
        ndecimals : int
            Number of decimals to round the chemical potentials, if None the numbers are not changed.

        Returns
        -------
        df : DataFrame
            DataFrame object.

        """
        res = self._get_res_dict_with_symbols(format_symbols)
        df = DataFrame(res)
        df = df.transpose()
        if format_compositions:
            new_index = []
            for string in df.index:
                new_string = format_composition(string,all_math=all_math)
                new_index.append(new_string)
            df.index = new_index
        if ndecimals:
            df = df.round(decimals=ndecimals)
        return df
        
    
    def get_latex_table(self,ndecimals=1):
        """
        Get string with formatted latex table of chemical potentials.
        """
        df = self.get_dataframe(format_symbols=True,ndecimals=ndecimals)
        table = df.to_latex(escape=False)
        return table
    
    
    def get_plot(self,elements,size=1,**kwargs):
        """
        Plot the stability diagram with the reservoir points.

        Parameters
        ----------
        elements : list
            List of strings with element symbols on the diagram axis.
        size : float
            Size of the points. The default is 1.
        **kwargs : dict
            Kwargs for the add_reservoirs function.

        Returns
        -------
        Matplotlib object.

        """
        from pynter.phase_diagram.plotter import PDPlotterAdder # import here to avoid circular import
        res = self.copy()
        if not res.pd:
            raise ValueError('PhaseDiagram object needs to be stored to plot the stability diagram')
        plt = PDHandler(res.pd).get_stability_diagram(elements)
        if not res.are_chempots_delta:
            res.set_to_referenced()
            
        plt = PDPlotterAdder(res.pd,size).add_reservoirs(res,elements,**kwargs)
        return plt
                
    
    def set_to_absolute(self):
        """
        Set reservoir dictionary to absolute values of chempots.
        """
        new_res_dict = self.get_absolute_res_dict()
        self.res_dict = new_res_dict
        self._are_chempots_delta = False
        return

        
    def set_to_referenced(self):
        """
        Set reservoir dictionary to referenced values of chempots.
        """
        new_res_dict = self.get_referenced_res_dict()
        self.res_dict = new_res_dict
        self._are_chempots_delta = True
        return    
    
                
    def _get_res_dict_with_symbols(self,format_symbols=False):
        """
        format_labels (bool): Format labels of element chempots in latex math format.
        """
        new_dict = {}
        for res,chempots in self.res_dict.items():
            new_dict[res] = {}
            chempots = chempots.mu #keep just dict for DataFrame
            for el in chempots:
                if format_symbols:
                    if self.are_chempots_delta:
                        label = '$\Delta \mu_{\text{%s}}$' %el
                    else:
                        label = '$\mu_{\text{%s}}$' %el
                else:
                    label = el
                new_dict[res][label] = chempots[el]
        return new_dict

    
class PressureReservoirs(Reservoirs):
    """
    Subclass of Reservoirs which contains temperature information. Useful for partial pressure analysis. 
    """
    def __init__(self,res_dict,temperature=None,phase_diagram=None,mu_refs=None,are_chempots_delta=False):
        super().__init__(res_dict,phase_diagram,mu_refs,are_chempots_delta)
        self.temperature = temperature
        self.pressures = list(self.res_dict.keys())
        
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.res_dict == other
        elif isinstance(other, PressureReservoirs):
            return self.res_dict == other.res_dict
        else:
            return False
        
    
    def as_dict(self):
        """
        Json-serializable dict representation of a PressureReservoirs object. The Pymatgen element 
        is expressed with the symbol of the element.

        Returns
        -------
        d : dict
            Json-serializable dict of a PressureReservoirs object.
        """
        d = {}
        d['@module'] = self.__class__.__module__
        d['@class'] = self.__class__.__name__
        d['res_dict'] = {r:mu.as_dict() for r,mu in self.res_dict.items()}
        d['temperature'] = self.temperature
        d['phase_diagram'] = self.pd.as_dict() if self.pd else None
        d['mu_refs'] = self.mu_refs.as_dict() if self.mu_refs else None
        d['are_chempots_delta'] = self.are_chempots_delta
        return d


    @classmethod
    def from_dict(cls,d):
        """
        Constructor of PressureReservoirs object from dictionary representation.
        
        Parameters
        ----------
        d : dict
        
        Returns
        -------
        PressureReservoirs object.
        
        """
        res_dict = {float(r):Chempots.from_dict(mu) for r,mu in d['res_dict'].items()}
        temperature = d['temperature'] if 'temperature' in d.keys() else None
        phase_diagram = PhaseDiagram.from_dict(d['phase_diagram']) if d['phase_diagram'] else None
        mu_refs = Chempots.from_dict(d['mu_refs'])  if d['mu_refs'] else None
        are_chempots_delta = d['are_chempots_delta']
            
        return cls(res_dict,temperature,phase_diagram,mu_refs,are_chempots_delta)


    @staticmethod
    def from_json(path_or_string):
        """
        Build PressureReservoirs object from json file or string.

        Parameters
        ----------
        path_or_string : str
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        PressureReservoir object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.load(path_or_string)
        return PressureReservoirs.from_dict(d)   
    
    
    



import numpy as np
from scipy.constants import k, h, pi, m_e
from scipy.optimize import bisect

from pymatgen.electronic_structure.dos import Dos, FermiDos, CompleteDos
from pymatgen.electronic_structure.core import Spin
from pymatgen.core.units import kb



def get_carrier_concentrations(dos, fermi_level, temperature, band_gap=None):
    """
    Get carrier concentrations by integrating density of states or using effective masses.
    
    Code from pymatgen.electronic_structure.dos.FermiDos.get_doping
    The function has been modified to return the absolute values of
    hole and electron concentrations.

    Calculate carrier concentrations at a given
    Fermi level and temperature. A simple Left Riemann sum is used for
    integrating the density of states over energy & equilibrium Fermi-Dirac
    distribution.

    Parameters
    ----------
    dos : dict or Dos
        Density of states to integrate. Can be provided as density of states D(E)
        or using effective masses.

        Format for effective masses is a dict with the following keys:

        - "m_eff_h" : holes effective mass in units of m_e (electron mass)
        - "m_eff_e" : electrons effective mass in units of m_h
        - `band_gap` : needs to be provided in arguments

        Format for explicit DOS (dictionary) with the following keys:

        - 'energies' : list or np.array with energy values
        - 'densities' : list or np.array with total density values
        - 'structure' : pymatgen Structure of the material, needed for DOS volume and charge normalization

        Alternatively, a pymatgen Dos object (Dos, CompleteDos, or FermiDos).

    fermi_level : float
        The Fermi level relative to the VBM in eV.
    temperature : float
        The temperature in Kelvin.
    band_gap : float
        The band gap in eV. If None is determined from the DOS.

    Returns:
    --------
        h : float
            Absolute value of hole concentration in 1/cm^3.
        n : float
            Absolute value of electron concentration in 1/cm^3.
    """   
    if type(dos) == dict:
        if 'energies' in dos and 'densities' in dos:
            E = dos['energies']
            D = dos['densities']
            structure = dos['structure']
            fdos = _get_fermidos_from_data(efermi=fermi_level,E=E,D=D,structure=structure,bandgap=band_gap)

        elif 'm_eff_e' in dos and 'm_eff_h' in dos:
            if not band_gap:
                raise ValueError('Band gap must be provided when computing DOS with effective masses')
            
            m_eff_h = dos['m_eff_h'] * m_e
            m_eff_e = dos['m_eff_e'] * m_e
            T = temperature

            E = fermi_level  # fermi_level referenced to VBM
            occupation_h = maxwell_boltzmann(E=E, T=T)
            h = get_dos_from_effective_mass(m_eff_h,T=T) * occupation_h
            
            E = band_gap - fermi_level  # fermi_level referenced to VBM
            occupation_e = maxwell_boltzmann(E=E, T=T) 
            n = get_dos_from_effective_mass(m_eff_e,T=T) * occupation_e
            return abs(h), abs(n)

    elif type(dos) in  (Dos,FermiDos,CompleteDos):
        fdos = FermiDos(dos=dos,bandgap=band_gap)
    else:
        raise ValueError('DOS must a dictionary (read function docs) or pymatgen Dos object')
    
    _,fdos_vbm = fdos.get_cbm_vbm()
    fermi_level = fdos_vbm + fermi_level

    cb_integral = np.sum(
        fdos.tdos[max(fdos.idx_mid_gap, fdos.idx_vbm + 1) :]
        * f0(fdos.energies[max(fdos.idx_mid_gap, fdos.idx_vbm + 1) :], fermi_level, temperature)
        * fdos.de[max(fdos.idx_mid_gap, fdos.idx_vbm + 1) :],
        axis=0,
    )
    vb_integral = np.sum(
        fdos.tdos[: min(fdos.idx_mid_gap, fdos.idx_cbm - 1) + 1]
        * f0(-fdos.energies[: min(fdos.idx_mid_gap, fdos.idx_cbm - 1) + 1], -fermi_level, temperature)
        * fdos.de[: min(fdos.idx_mid_gap, fdos.idx_cbm - 1) + 1],
        axis=0,
    )
    h = (vb_integral) / (fdos.volume * fdos.A_to_cm ** 3) 
    n = -1*(cb_integral) / (fdos.volume * fdos.A_to_cm ** 3)
    
    return abs(h), abs(n)


def solve_intrinsic_fermi_level(dos,temperature,band_gap,xtol=1e-05):
    """
    Find Fermi level by solving the charge neutrality condition.

    Parameters
    ----------
    dos : dict or Dos
        Density of states to integrate. Can be provided as density of states D(E)
        or using effective masses.

        Format for effective masses is a dict with the following keys:

        - "m_eff_h" : holes effective mass in units of m_e (electron mass)
        - "m_eff_e" : electrons effective mass in units of m_h
        - `band_gap` : needs to be provided in arguments

        Format for explicit DOS (dictionary) with the following keys:

        - 'energies' : list or np.array with energy values
        - 'densities' : list or np.array with total density values
        - 'structure' : pymatgen Structure of the material, needed for DOS volume and charge normalization

        Alternatively, a pymatgen Dos object (Dos, CompleteDos, or FermiDos).

    temperature : float
        The temperature in Kelvin.
    band_gap : float
        The band gap in eV. If None is determined from the DOS.
    xtol : float
        Tolerance for `bisect` method (scipy).

    Returns
    -------
    fermi_level : float
        Fermi level dictated by charge neutrality.

    """
    def _get_total_q(ef):
        h,n = get_carrier_concentrations(dos=dos,fermi_level=ef,temperature=temperature,band_gap=band_gap)
        return h - n
    
    return bisect(_get_total_q, -1., band_gap + 1.,xtol=xtol)


def f0(E, fermi, T):
    """
    Fermi-Dirac distribution function.

    Parameters
    ----------
    E : float
        Energy in eV.
    fermi : float
        Fermi level in eV.
    T : float
        Temperature in kelvin.

    Returns
    -------
    occupation : float
        Fermi-Dirac occupation probability at energy E.
    """
    from scipy.special import expit
    exponent = (E - fermi) / (kb * T)
    return expit(-exponent) 

def maxwell_boltzmann(E, T, clip=True):
    """
    Maxwell-Boltzmann distribution function.
    Can be clipped to prevent divergence

    Parameters
    ----------
    E : float
        Energy in eV.
    fermi : float
        Fermi level in eV.
    T : float
        Temperature in kelvin.
    clip : bool
        Clip exponential factor to 700 to avoid divergence, returns 1e304.

    Returns
    -------
    occupation : float
        Maxwell-Boltzmann occupation probability at energy E.

    """
    factor = -E / (kb*T)
    if clip:
        factor = np.clip(factor,None,700)
    return np.exp(factor)


def get_dos_from_effective_mass(m_eff, T):
    """
    Calculate the effective density of states (N_c or N_v)
    for a non-degenerate semiconductor.

    Parameters
    ----------
    m_eff : float
        Effective mass (in units of kg)
    T : float
        Temperature (in Kelvin)

    Returns
    -------
    N : float
        Density of states (in cm^-3).

    """
    from scipy.constants import k, h, pi, m_e
    return 2 * ((2 * pi * m_eff * k * T) / (h**2))**(3/2) *1e-06  # units of cm^-3



def _get_fermidos_from_data(efermi, E, D, structure, bandgap=None):
    """
    Get `FermiDos` object from energies, densities and Structure. 
    """
    if type(E) == list:
        E = np.array(E)
    if type(D) == list:
        D = np.array(D)

    dos = Dos(efermi=efermi,energies=E,densities={Spin.up:D})
    return FermiDos(dos=dos,structure=structure,bandgap=bandgap)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 12 11:59:40 2023

@author: villa
"""
import numpy as np
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.composition import Composition


from defermi.defects import Vacancy, Interstitial, Substitution, Polaron, DefectComplex

from defermi.testing.core import DefermiTest
from defermi.testing.defects import DefectTest


bulk_structure = DefermiTest().structure.copy()
bulk_structure.make_supercell(3)


class TestDefect(DefermiTest):

    def test_interstitial(self):
        structure = bulk_structure.copy()
        frac_coords = np.array([0.66666667, 0.5, 0.5])
        interstitial_site = PeriodicSite('Si', frac_coords, structure.lattice)
        inter = Interstitial(
                           specie='Si',
                           defect_site=interstitial_site,
                           bulk_structure=structure,
                           charge=0,
                           multiplicity=108,
                           label='test') 
        
        assert inter.specie == 'Si'
        assert inter.type == 'Interstitial'
        assert inter.defect_site_index == 54
        assert inter.defect_composition == Composition('Si55')
        assert inter.defect_structure.composition == inter.defect_composition
        assert inter.name == 'Int_Si(test)'
        assert inter.charge == 0
        assert inter.delta_atoms == {'Si': 1}
        assert inter.symbol == '$Si_i$(test)'
        assert inter.symbol_with_charge == '$Si_i$(test)$^{\\;0}$'
        assert inter.symbol_with_charge_kv == '$Si_i$(test)$^{x}$'
    
    
    def test_polaron(self):
        structure = bulk_structure.copy()
        site = structure[0]
        pol = Polaron(
                    specie='Si',
                    defect_site=site,
                    bulk_structure=structure,
                    charge=1,
                    multiplicity=1,
                    label='test') 

        assert pol.specie == 'Si'
        assert pol.type == 'Polaron'
        assert pol.defect_site_index == 0
        assert pol.defect_composition == Composition('Si54')
        assert pol.defect_structure.composition == pol.defect_composition
        assert pol.charge == 1
        assert pol.delta_atoms == {}
        assert pol.get_multiplicity() == 54
        pol.set_multiplicity(54)
        assert pol.multiplicity == 54
        assert pol.name == 'Pol_Si(test)'
        assert pol.symbol == '$Si_{Si}$(test)'
        assert pol.symbol_with_charge =='$Si_{Si}$(test)$^{+1}$'
        assert pol.symbol_with_charge_kv == '$Si_{Si}$(test)$^{°}$'
    
    
    def test_substitution(self):
        structure = bulk_structure.copy()
        site = structure[0]
        defect_site = PeriodicSite('P',site.frac_coords,site.lattice)
        sub = Substitution(
                    specie='P',
                    bulk_specie='Si',
                    defect_site=site,
                    bulk_structure=structure,
                    charge=1,
                    multiplicity=1,
                    label='test') 
 
        assert sub.specie == 'P'
        assert sub.type == 'Substitution'
        assert sub.defect_site_index == 0
        assert sub.defect_composition == Composition('Si53P1')
        assert sub.defect_structure.composition == sub.defect_composition
        assert sub.name == 'Sub_P_on_Si(test)'
        assert sub.charge == 1
        assert sub.delta_atoms == {'P': 1, 'Si': -1}
        assert sub.site_in_bulk == site
        assert sub.get_multiplicity() == 54
        sub.set_multiplicity(54)
        assert sub.multiplicity == 54
        assert sub.symbol == '$P_{Si}$(test)'
        assert sub.symbol_with_charge == '$P_{Si}$(test)$^{+1}$'
        assert sub.symbol_with_charge_kv == '$P_{Si}$(test)$^{°}$'
        
    
    def test_vacancy(self):
        structure = bulk_structure.copy()
        site = structure[0]
        vac = Vacancy(
                           specie='Si',
                           defect_site=site,
                           bulk_structure=structure,
                           charge=0,
                           multiplicity=1,
                           label='test') 

        assert vac.specie == 'Si'
        assert vac.type == 'Vacancy'
        assert vac.defect_site_index == 0
        assert vac.defect_composition == Composition('Si53')
        assert vac.defect_structure.composition == vac.defect_composition
        assert vac.charge == 0
        assert vac.delta_atoms == {'Si':-1}
        assert vac.get_multiplicity() == 54
        vac.set_multiplicity(54)
        assert vac.multiplicity == 54
        assert vac.name == 'Vac_Si(test)'
        assert vac.symbol == '$V_{Si}$(test)'
        assert vac.symbol_with_charge =='$V_{Si}$(test)$^{\\;0}$'
        assert vac.symbol_with_charge_kv == '$V_{Si}$(test)$^{x}$'
        
        
    def test_defect_complex(self):
        structure = bulk_structure.copy()
        vac_site = structure[0]
        vac = Vacancy(
                           specie='Si',
                           defect_site=vac_site,
                           bulk_structure=structure,
                           charge=0,
                           multiplicity=54,
                           label='vac') 

        sub_site = structure[1]
        defect_site = PeriodicSite('P',sub_site.frac_coords,sub_site.lattice)
        sub = Substitution(
                           specie='P',
                           bulk_specie='Si',
                           defect_site=sub_site,
                           bulk_structure=structure,
                           charge=1,
                           multiplicity=54,
                           label='sub') 

        defects = [vac,sub]
        comp = DefectComplex(
                            defects=defects,
                            bulk_structure=structure,
                            multiplicity=54*4,
                            charge=1,
                            label='test')
        assert comp.type == 'DefectComplex'
        assert comp.name == 'Vac_Si(vac)-Sub_P_on_Si(sub)(test)'
        assert comp.defect_composition == Composition('Si52P1')
        assert comp.charge == 1
        assert comp.delta_atoms == {'Si': -2, 'P': 1}
        assert comp.multiplicity == 216
        assert comp.symbol == '$V_{Si}$(vac)-$P_{Si}$(sub)(test)'
        assert comp.symbol_with_charge == '$V_{Si}$(vac)-$P_{Si}$(sub)(test)$^{+1}$'
        assert comp.symbol_with_charge_kv == '$V_{Si}$(vac)-$P_{Si}$(sub)(test)$^{°}$'
    
    

    




















